/*
** dump.c - utility to display contents of slave windows on an ADM-XRC
**          series card.
**
** (C) Copyright 2009-2010 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcslen(x) strlen(x)
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _gettc getc
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MAX_TOKEN_LENGTH (128)

typedef enum _TokenSubstate {
  NotInToken,
  InTokenDecimal,
  InTokenHexadecimal,
  InTokenInvalid
} TokenSubstate;

typedef struct _TokenState {
  unsigned int position;
  TokenSubstate substate;
  unsigned long long value;
  TCHAR tokenBuf[MAX_TOKEN_LENGTH];
} TokenState;

static const TokenState g_tokenStateInit = {
  0, NotInToken
};

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

static size_t
remaining(
  const TCHAR* pPos,
  const TCHAR* line,
  size_t max)
{
  size_t n = pPos - line;
  if (n > max) {
    return 0;
  } else {
     return max - n;
  }
}

static void
dumpMemory(
	unsigned int indent,
	unsigned int nColumn,
	unsigned int radix,
  bool_t bBigEndian,
	bool_t bAlign,
	const void* pBuffer,
	size_t n)
{
	uint64_t address, limit, addressCol, m;
	unsigned int addressLo, columnMask;
	uint8_t* p;
	uint8_t* p8;
	TCHAR line[257];
	TCHAR* pPos;
	uint8_t columnBuf[256];
	uint8_t val8;
	uint16_t val16;
	uint32_t val32;
	uint64_t val64;
	unsigned int i, j, k, chunk;
  unsigned int rowBytes;

	assert(sizeof(void*) == 4 || sizeof(void*) == 8);
	if (indent > 128) {
		_tprintf(_T("*** dumpMemory: invalid value '%lu' for indent\n"), (unsigned long)indent);
		return;
	}
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    _tprintf(_T("*** dumpMemory: invalid value '%lu' for nColumn\n"), (unsigned long)nColumn);
    return;
  }
	if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
		_tprintf(_T("*** dumpMemory: invalid value '%lu' for radix\n"), (unsigned long)radix);
		return;
	}

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
	columnMask = rowBytes - 1U;
	address = (uint64_t)(uintptr_t)pBuffer;
	limit = address + n;
	if (bAlign) {
		addressCol = address & ~(uint64_t)columnMask;
		addressLo = ((unsigned int)address) & columnMask;
	} else {
		addressCol = address;
		addressLo = 0;
	}

	m = limit - address;
	p = (uint8_t*)pBuffer;

	pPos = line;
	pPos[0] = _T('\0');
  _tcsnset_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T(' '), indent);
	pPos += indent;
	_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("Dump of memory at 0x%p + %lu(0x%lx) bytes:\n"),
		pBuffer, (unsigned long)n, (unsigned long)n);
	_tprintf(_T("%s"), line);

	pPos = line;
	if (sizeof(void*) == 4) {
    _tcsnset_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T(' '), indent + 12);
		pPos += indent + 12;
	} else {
    _tcsnset_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T(' '), indent + 23);
		pPos += indent + 23;
	}
	for (i = 0; i < rowBytes; i += radix) {
		switch (radix) {
			case 0x1U:
				_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("%02lx "), (unsigned long)i);
				pPos += 3;
				break;

			case 0x2U:
				_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("  %02lx "), (unsigned long)i);
				pPos += 5;
				break;

			case 0x4U:
				_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("      %02lx "), (unsigned long)i);
				pPos += 9;
				break;

			case 0x8U:
				_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("               %02lx "), (unsigned long)i);
				pPos += 18;
				break;
		}
	}
	*pPos++ = _T('\n');
	*pPos++ = _T('\0');
	_tprintf(_T("%s"), line);

	while (m) {
		chunk = rowBytes - addressLo;
		if (chunk > m) {
			chunk = (unsigned int)m;
		}

		memset(columnBuf, 0, rowBytes);
		memcpy(columnBuf + addressLo, p, chunk);

		p8 = (uint8_t*)columnBuf;

		pPos = line;
    _tcsnset_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T(' '), indent);
		pPos += indent;
		if (sizeof(void*) == 4) {
			_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("0x%08lx: "), (unsigned long)(uint32_t)addressCol);
			pPos += 12;
		} else {
			_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("0x%08lx_0x%08lx: "),
				(unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
			pPos += 23;
		}

		j = addressLo + chunk;
		for (i = 0; i < rowBytes; i += radix) {
			switch (radix) {
				case 0x1U:
					val8 = *p8++;
					if (i >= addressLo && i < j) {
						_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("%02lx"), (unsigned long)val8);
						pPos += 2;
					} else {
						*pPos++ = _T('.');
						*pPos++ = _T('.');
					}
					break;

				case 0x2U:
          if (bBigEndian) {
					  val16 = (((uint16_t)p8[0]) << 8U) | (((uint16_t)p8[1]) << 0U);
          } else {
					  val16 = (((uint16_t)p8[0]) << 0U) | (((uint16_t)p8[1]) << 8U);
          }
          p8 += 2;
					for (k = 2; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("%02lx"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
							pPos += 2;
						} else {
							*pPos++ = _T('.');
							*pPos++ = _T('.');
						}
						if (k == 3) {
							*pPos++ = _T('_');
						}
					}
					break;

				case 0x4U:
          if (bBigEndian) {
  					val32 = (((uint32_t)p8[0]) << 24U) | (((uint32_t)p8[1]) << 16U) | (((uint32_t)p8[2]) <<  8U) | (((uint32_t)p8[3]) <<  0U);
          } else {
	  				val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
          }
          p8 += 4;
					for (k = 4; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("%02lx"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
							pPos += 2;
						} else {
							*pPos++ = _T('.');
							*pPos++ = _T('.');
						}
					}
					break;

				case 0x8U:
          if (bBigEndian) {
  					val64  = (((uint64_t)p8[0]) << 56U) | (((uint64_t)p8[1]) << 48U) | (((uint64_t)p8[2]) << 40U) | (((uint64_t)p8[3]) << 32U);
	  				val64 |= (((uint64_t)p8[4]) << 24U) | (((uint64_t)p8[5]) << 16U) | (((uint64_t)p8[6]) <<  8U) | (((uint64_t)p8[7]) <<  0U);
          } else {
  					val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
	  				val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
          }
          p8 += 8;
					for (k = 8; k != 0;) {
						k--;
						if (i + k >= addressLo && i + k < j) {
							_stprintf_s(pPos, remaining(pPos, line, ARRAY_LENGTH(line)), _T("%02lx"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
							pPos += 2;
						} else {
							*pPos++ = _T('.');
							*pPos++ = _T('.');
						}
						if (k == 4) {
							*pPos++ = _T('_');
						}
					}
					break;
			}
			*pPos++ = _T(' ');
		}

		p8 = (uint8_t*)columnBuf;
		for (i = 0; i < rowBytes; i++) {
			char c = (char)*p8++;

			if (_istprint(c)) {
				*pPos++ = c;
			} else {
				*pPos++ = _T('.');
			}
		}
		*pPos++ = _T('\n');
		*pPos++ = _T('\0');
    _tprintf(_T("%s"), line);
		addressLo = 0;
		addressCol += rowBytes;
		m -= chunk;
		p += chunk;
	}
}

/* Returns TRUE if a token has just been fully parsed */
static bool_t
pumpTokenState(
  TokenState* pState,
  TCHAR c,
  bool_t bHex)
{
  TokenSubstate nextState = pState->substate;
  bool_t bParsed = FALSE, bStore = FALSE;

  if (c == _T('\0')) {
    c = _T(' ');
  } else {
    c = tolower(c);
  }
  switch (pState->substate) {
  case NotInToken:
    if (!isspace(c)) {
      if (c >= _T('a') && c <= _T('f')) {
        nextState = InTokenHexadecimal;
        bStore = TRUE;
      } else {
        if (c >= _T('0') && c <= _T('9')) {
          nextState = bHex ? InTokenHexadecimal : InTokenDecimal;
          bStore = TRUE;
        } else {
          nextState = InTokenInvalid;
        }
      }
    }
    break;

  case InTokenDecimal:
    if (tolower(c) == _T('x') && pState->position == 1) {
      nextState = InTokenHexadecimal;
      pState->position = 0;
    } else if (c >= _T('a') && c <= _T('f')) {
      nextState = InTokenHexadecimal;
      bStore = TRUE;
    } else if (c >= _T('0') && c <= _T('9')) {
      nextState = InTokenDecimal;
      bStore = TRUE;
    } else if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      nextState = InTokenInvalid;
    }
    break;

  case InTokenHexadecimal:
    if ((c >= _T('a') && c <= _T('f')) || (c >= _T('0') && c <= _T('9'))) {
      nextState = InTokenHexadecimal;
      bStore = TRUE;
    } else if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      pState->substate = InTokenInvalid;
    }
    break;

  case InTokenInvalid:
    if (isspace(c)) {
      nextState = NotInToken;
      bParsed = TRUE;
    } else {
      nextState = InTokenInvalid;
    }
    break;
  }

  if (bStore) {
    if (pState->position < ARRAY_LENGTH(pState->tokenBuf) - 1U) {
      pState->tokenBuf[pState->position] = c;
      pState->position++;
    }
  }

  if (bParsed) {
    pState->tokenBuf[pState->position] = _T('\0');
    if (pState->position == 0) {
      pState->value = 0;
    } else {
      if (pState->substate == InTokenHexadecimal) {
        _stscanf_s(pState->tokenBuf, _T("%llx"), &pState->value);
      } else if (pState->substate == InTokenDecimal) {
        _stscanf_s(pState->tokenBuf, _T("%llu"), &pState->value);
      } else {
        pState->value = 0;
      }
      pState->position = 0;
    }
  }
  pState->substate = nextState;

  return bParsed;
}

static void
writeValueToWindowUnaligned(
  ADMXRC3_HANDLE hCard,
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t val64)
{
  switch (width) {
  case 1U:
    *p++ = (uint8_t)val64;
    break;

  case 2U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
    }
    break;

  case 4U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 24U);
    }
    break;

  case 8U:
    if (bBigEndian) {
      *p++ = (uint8_t)(val64 >> 56U);
      *p++ = (uint8_t)(val64 >> 48U);
      *p++ = (uint8_t)(val64 >> 40U);
      *p++ = (uint8_t)(val64 >> 32U);
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 0U);
    } else {
      *p++ = (uint8_t)(val64 >> 0U);
      *p++ = (uint8_t)(val64 >> 8U);
      *p++ = (uint8_t)(val64 >> 16U);
      *p++ = (uint8_t)(val64 >> 24U);
      *p++ = (uint8_t)(val64 >> 32U);
      *p++ = (uint8_t)(val64 >> 40U);
      *p++ = (uint8_t)(val64 >> 48U);
      *p++ = (uint8_t)(val64 >> 56U);
    }
    break;

  default:
    assert(FALSE);
    break;
  }
}

static void
writeValueToWindow(
  ADMXRC3_HANDLE hCard,
  uint8_t* p,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t val64)
{
  uint16_t* p16;
  uint32_t* p32;
  uint64_t* p64;

  if ((uintptr_t)p & (width - 1U)) {
    writeValueToWindowUnaligned(hCard, p, width, bBigEndian, val64);
    return;
  }

  switch (width) {
  case 1U:
    *p++ = (uint8_t)val64;
    break;

  case 2U:
    p16 = (uint16_t*)p;
    if (bBigEndian) {
      *p16 = (uint16_t)cpu_to_be16((uint16_t)val64);
    } else {
      *p16 = (uint16_t)cpu_to_le16((uint16_t)val64);
    }
    break;

  case 4U:
    p32 = (uint32_t*)p;
    if (bBigEndian) {
      *p32 = (uint32_t)cpu_to_be32((uint32_t)val64);
    } else {
      *p32 = (uint32_t)cpu_to_le32((uint32_t)val64);
    }
    break;

  case 8U:
    p64 = (uint64_t*)p;
    if (bBigEndian) {
      *p64 = (uint64_t)cpu_to_be64(val64);
    } else {
      *p64 = (uint64_t)cpu_to_le64(val64);
    }
    break;

  default:
    assert(FALSE);
    break;
  }
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  uint8_t* pMapped = NULL;
  Option options[4];
  Arguments arguments = g_initArguments;
	OptionType paramType[4];
	TCHAR* paramInfo[4];
	TCHAR* paramMeaning[4];
  bool_t bBigEndian, bHex, bWrite;
  unsigned int n, width, index, serial, windowIndex;
  uint64_t offset;
  TCHAR* pCommand;
  int ret = 0;

	options[0].key = _T("index");
	options[0].type = OptionUInt;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
	options[2].key = _T("be");
	options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("TRUE => interpret values as being decimal");
	options[3].key = _T("hex");
	options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
	options[3].meaning = _T("TRUE => interpret values as being hexadecimal");
	arguments.nOption = 4;
	arguments.option = options;
	paramType[0] = OptionString;
	paramInfo[0] = _T("command");
  paramMeaning[0] = _T("Command to perform: rb, rw, rd, rq, wb, ww, wd or wq");
	paramType[1] = OptionUInt;
	paramInfo[1] = _T("window");
	paramMeaning[1] = _T("Index of window to access");
	paramType[2] = OptionUInt64;
	paramInfo[2] = _T("offset");
	paramMeaning[2] = _T("Offset in window");
	paramType[3] = OptionUInt;
	paramInfo[3] = _T("length");
	paramMeaning[3] = _T("Number of bytes to read or write");
	arguments.nParamType = 4;
	arguments.minNParam = 3;
	arguments.paramType = paramType;
	arguments.paramInfo = paramInfo;
	arguments.paramMeaning = paramMeaning;
	
  if (sampleParseCommandLine(_T("dump"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bBigEndian = options[2].value.booleanVal;
  bHex = options[3].value.booleanVal;
  pCommand = arguments.param[0].pStringVal;
  windowIndex = arguments.param[1].uintVal;
  offset = arguments.param[2].uint64Val;

  if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    bWrite = FALSE;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    bWrite = FALSE;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    bWrite = FALSE;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    bWrite = FALSE;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    bWrite = TRUE;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    bWrite = TRUE;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    bWrite = TRUE;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    bWrite = TRUE;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = -1;
    goto done;
  }

  if (arguments.nParam < 4) {
    n = width;
  } else {
    n = arguments.param[3].uintVal;
  }
  if (n % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n"), n, n, width);
    n &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%x)\n"), n, n);
  }

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_MapWindow(hCard, windowIndex, offset, n, (void**)&pMapped);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map window %u offset 0x%llx + 0x%x: %s\n"),
      windowIndex, (unsigned long long)offset, n, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  _tprintf(_T("Window %u offset 0x%llx mapped @ 0x%p\n"), windowIndex, (unsigned long long)offset, pMapped);

  if (bWrite) {
    TokenState state = g_tokenStateInit;
    uint8_t* p = pMapped;
    TCHAR* pArg;
    TCHAR c;
    size_t j;
    int i;

    /* Read tokens from remaining command line arguments */
    i = 4;
    while (i < arguments.nParam && n > 0) {
      state = g_tokenStateInit;
      pArg = arguments.param[i].pStringVal;
      j = 0;
      while (1) {
        c = pArg[j];
        j++;
        if (pumpTokenState(&state, c, bHex)) {
          _tprintf(_T("0x%llX: 0x%llX\n"), (unsigned long long)offset, state.value);
          writeValueToWindow(hCard, p, width, bBigEndian, (uint64_t)state.value);
          p += width;
          n -= width;
          offset += width;
        }
        if (_T('\0') == c) {
          break;
        }
      }
      i++;
    }
    if (n == 0) {
      goto done;
    }
    /* Read tokens from stdin to satisfy remaining byte count (if nonzero) */
    state = g_tokenStateInit;
    _tprintf(_T("0x%llX: "), (unsigned long long)offset);
    while (n > 0) {
      c = _gettc(stdin);
      if (c == EOF) {
        goto done;
      }
      if (pumpTokenState(&state, c, bHex)) {
        writeValueToWindow(hCard, p, width, bBigEndian, (uint64_t)state.value);
        p += width;
        n -= width;
        offset += width;
        if (n > 0) {
          _tprintf(_T("0x%llX: "), (unsigned long long)offset);
        }
      }
    }
  } else {
    dumpMemory(0, 16 / width, width, bBigEndian, TRUE, pMapped, n);
  }

done:
  if (NULL != pMapped) {
    if (ADMXRC3_SUCCESS != ADMXRC3_UnmapWindow(hCard, pMapped)) {
      _tprintf(_T("+++ Failed to unmap window at 0x%p: %s\n"), pMapped, ADMXRC3_GetStatusString(status, TRUE));
    }
    pMapped = NULL;
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
