/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADMXRC3 API via Win32 events.
**
** (C) Copyright 2009-2010 Alpha Data
*/

#include <windows.h>
#include <stdio.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

/*
** FPGA register addresses on local bus. These are also the offsets from the
** base of space 0, since space 0 begins at absolute local bus address 0x0.
*/
#define REGADDR_ISET 	(0xC0U)
#define REGADDR_ISTAT (0xC4U)
#define REGADDR_IMASK (0xC8U)
#define REGADDR_IARM	(0xCCU)

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  HANDLE hEvent;
  BOOL bTerminate;
  unsigned long signalled;
} ThreadContext;

static DWORD WINAPI
interruptThread(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
	DWORD result;
	DWORD istat;
	DWORD dummy = 0; /* value doesn't matter, but avoid compiler whine */

	_tprintf(_T("Interrupt thread started\n"));
	
	while (!pContext->bTerminate) {
		result = WaitForSingleObject(pContext->hEvent, INFINITE);
		if (result != WAIT_OBJECT_0) {
      DWORD lastError = GetLastError();

			_tprintf(_T("*** Interrupt thread failed to wait for event (Win32 error %lu(0x%lX))\n"),
        (unsigned long)lastError, (unsigned long)lastError);
			break;
		}

		if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      ADMXRC3_Read(pContext->hCard,	0, 0, REGADDR_ISTAT, sizeof(istat), &istat);

      /* Clear the pending FPGA interrupts */
      ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_ISTAT, sizeof(istat), &istat);

      /* Rearm the 'FINTI_L' line driven by the FPGA by forcing it high for 1 cycle */
      ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_IARM, sizeof(dummy), &dummy);
    }
	}

	return 0;
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
	const double clockFreq = 33.0e6;
	int ret = 0;
  unsigned int index, serial;
	ADMXRC3_STATUS status;
	ADMXRC3_CARD_INFO cardInfo;
	unsigned long numInterruptGenerated;
	const TCHAR* pFilename;
  Arguments arguments = g_initArguments;
	Option options[2];
	DWORD val32;
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    NULL, /* hEvent */
    FALSE, /* bTerminate */
    0 /* signalled */
  };
  BOOLEAN bEventRegistered = FALSE;
  HANDLE hInterruptThread = NULL;
  DWORD interruptThreadId;
  BOOLEAN bInterruptThreadRunning = FALSE;
  int ch;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
	arguments.nOption = 2;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("itest"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
	if (status != ADMXRC3_SUCCESS) {
		_tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		ret = -1;
		goto done;
	}
	
  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  pFilename = sampleGetFpgaFilename(threadContext.hCard, 0, _T("uber"), _T("uber"), &arguments);
  if (NULL == pFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, pFilename);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to load the bitstream '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

	/* Create an auto-reset event */
	threadContext.hEvent = CreateEvent(NULL, FALSE, FALSE,	NULL);
	if (threadContext.hEvent == NULL) {
		_tprintf(_T("*** Failed to create a Win32 event\n"));
		ret = -1;
		goto done;
	}

	/*
	** Register the event with the card, so that when an FPGA interrupt occurs,
	** the event will be signalled.
	*/
	status = ADMXRC3_RegisterWin32Event(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.hEvent);
	if (status != ADMXRC3_SUCCESS) {
		_tprintf(_T("*** Failed to register Win32 event for FPGA interrupt: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		ret = -1;
		goto done;
	}
	bEventRegistered = TRUE;

	/* Start up the thread that waits for interrupts. */
	threadContext.signalled = 0;
	threadContext.bTerminate = FALSE;
	hInterruptThread = CreateThread(NULL, 0, interruptThread, (void*)&threadContext, 0, &interruptThreadId);
	if (hInterruptThread == NULL) {
		_tprintf(_T("*** Failed to start interrupt thread\n"));
		ret = -1;
		goto done;
	}
	bInterruptThreadRunning = TRUE;

	/* Enable interrupts in the FPGA */
	val32 = cpu_to_le32(0x0U);
	status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_IMASK, sizeof(val32), &val32);
	if (status != ADMXRC3_SUCCESS) {
		_tprintf(_T("*** Write to FPGA_IMASK register failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
		ret = -1;
		goto done;
	}

	/*
	** Allow the user to generate interrupts by pressing the enter key.
	** Ctrl-Z or entering 'q' / 'Q' exits this loop.
	*/
	numInterruptGenerated = 0;
	do {
		_tprintf(_T("Enter 'q' to quit, or anything else to generate an interrupt:\n"));
		ch = getchar();
		if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_ISET, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Write to FPGA_TEST register failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      }

      numInterruptGenerated++;
    }
	} while (EOF != ch);

	_tprintf(_T("Generated %lu interrupts\n"), numInterruptGenerated);
	_tprintf(_T("Interrupt thread saw %lu interrupt(s)\n"), threadContext.signalled);

done:
	if (bInterruptThreadRunning) {
		/* Force the interrupt thread to terminate */
		threadContext.bTerminate = TRUE;
		SetEvent(threadContext.hEvent);
		/* Wait for the interrupt thread to terminate */
		WaitForSingleObject(hInterruptThread, 1000);
	}

	if (threadContext.hEvent != NULL) {
		if (bEventRegistered) {
			/*
			** Unregister the event, so that it will no longer be signalled
			** on an FPGA interrupt.
			*/
			status = ADMXRC3_UnregisterWin32Event(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.hEvent);
			if (status != ADMXRC3_SUCCESS) {
				_tprintf(_T("*** Failed to unregister interrupt event: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
			}
		}
		CloseHandle(threadContext.hEvent);
	}

  if (threadContext.hCard != NULL) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
