/*
** monitor.c - Trivial example showing how to read sensor values from
**             the system monitoring hardware on Gen 3 Alpha Data
**             reconfigurable computing hardware.
**
** (C) Copyright 2010 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
# define _stprintf_s snprintf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_time.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FORMFEED (12)

static const struct {
  ADMXRC3_MODEL_TYPE model;
  const TCHAR* pName;
} models[] = {
  { ADMXRC3_MODEL_GENERIC, _T("Generic reconfigurable computing card") },
  { ADMXRC3_MODEL_ADMXRC2, _T("ADM-XRC-II") },
  { ADMXRC3_MODEL_ADPEXRC5T, _T("ADPe-XRC-5T") },
  { ADMXRC3_MODEL_ADMXRC6TL, _T("ADM-XRC-6TL") },
  { ADMXRC3_MODEL_ADMXRC6T1, _T("ADM-XRC-6T1") }
};

static const TCHAR*
getModelName(
  ADMXRC3_MODEL_TYPE modelType)
{
  int i;

  for (i = 0; i < ARRAY_LENGTH(models); i++) {
    if (models[i].model == modelType) {
      return models[i].pName;
    }
  }

  return NULL;
}

static const TCHAR*
getUnit(
  ADMXRC3_UNIT_TYPE unit)
{
  switch (unit) {
  case ADMXRC3_UNIT_NONE:
    return _T("");

  case ADMXRC3_UNIT_A:
    return _T("A");

  case ADMXRC3_UNIT_V:
    return _T("V");

  case ADMXRC3_UNIT_C:
    return _T("deg C");

  case ADMXRC3_UNIT_HZ:
    return _T("Hz");

  case ADMXRC3_UNIT_RPM:
    return _T("RPM");

  default:
    return _T("(unknown unit)");
  }
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  ADMXRC3_STATUS status;
  unsigned int index, serial, numRep, period;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_CARD_INFOEX cardInfoEx;
  ADMXRC3_SENSOR_INFO sensorInfo;
  TCHAR modelName[80];
  Option options[4];
  Arguments arguments = g_initArguments;
  unsigned int i, rep;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("repeat");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("number of repetitions (0 => infinite)");
  options[3].key = _T("period");
  options[3].type = OptionUInt;
  options[3].def.uintVal = 1;
  options[3].meaning = _T("update period, in seconds");
  arguments.nOption = 4;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("monitor"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  numRep = options[2].value.booleanVal;
  period = options[3].value.booleanVal;
  if (period < 1U) {
    period = 1U;
  }
  if (period > 100U) {
    period = 100U;
  }

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, TRUE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfoEx(hCard, &cardInfoEx);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  if (NULL != getModelName(cardInfoEx.Model)) {
    _tcscpy_s(modelName, ARRAY_LENGTH(modelName), getModelName(cardInfoEx.Model));
  } else {
    _stprintf_s(modelName, ARRAY_LENGTH(modelName), _T("unknown (0x%lX)"), (unsigned long)cardInfoEx.Model);
  }

  for (rep = 0; numRep == 0 || rep < numRep; rep++) {
    ADMXRC3_SENSOR_VALUE sensorValue;

    _tprintf(_T("Model:                                    %lu (0x%lX) => %s \n"), (unsigned long)cardInfoEx.Model, (unsigned long)cardInfoEx.Model, modelName);
    _tprintf(_T("Serial number:                            %lu (0x%lX)\n"), (unsigned long)cardInfoEx.SerialNumber, (unsigned long)cardInfoEx.SerialNumber);
    _tprintf(_T("Number of sensors:                        %lu\n"), (unsigned long)cardInfoEx.NumSensor);

    if (0 == cardInfoEx.NumSensor) {
      _tprintf(_T("*** No sensors to monitor. Exiting.\n"));
      ret = -1;
      goto done;
    }

    for (i = 0; i < cardInfoEx.NumSensor; i++) {
      _tprintf(_T("  Sensor %2u "), i);
      status = ADMXRC3_GetSensorInfo(hCard, i, &sensorInfo);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("(Failed to get info: %s)\n"), ADMXRC3_GetStatusString(status, TRUE));
        continue;
      }

      status = ADMXRC3_ReadSensor(hCard, i, &sensorValue);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("%28s: (Failed to read sensor: %s)\n"), sensorInfo.Description, ADMXRC3_GetStatusString(status, TRUE));
        continue;
      }

      switch (sensorInfo.DataType) {
      case ADMXRC3_DATA_BOOL:
        assert(sensorInfo.Exponent == 0 && sensorInfo.Unit == ADMXRC3_UNIT_NONE);
        _tprintf(_T("%28s: %s\n"), sensorInfo.Description, sensorValue.Bool ? "TRUE" : "FALSE");
        break;

      case ADMXRC3_DATA_DOUBLE:
        _tprintf(_T("%28s: %f"), sensorInfo.Description, sensorValue.Double);
        break;

      case ADMXRC3_DATA_INT32:
        _tprintf(_T("%28s: %ld"), sensorInfo.Description, (long)sensorValue.Int32);
        break;

      case ADMXRC3_DATA_UINT32:
        _tprintf(_T("%28s: %lu"), sensorInfo.Description, (unsigned long)sensorValue.UInt32);
        break;

      default:
        /* Unknown type; can't interpret */
        _tprintf(_T("%28s: ?"), sensorInfo.Description);
        break;
      }
      if (sensorInfo.DataType != ADMXRC3_DATA_BOOL) {
        if (sensorInfo.Exponent) {
          _tprintf(_T(" * 10e%d %s\n"), sensorInfo.Exponent, getUnit(sensorInfo.Unit));
        } else {
          _tprintf(_T(" %s\n"), getUnit(sensorInfo.Unit));
        }
      }
    }

    _tprintf(_T("\n"));
    sleepMs(period * 1000);
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
