/*
** info.c - displays information about an ADM-XRC Gen 3 card.
**
** (C) Copyright 2010 Alpha Data
*/

#include <vxWorks.h>
#include <stdio.h>
#include <string.h>

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#include <admxrc3.h>
#include "sdk_common.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const struct {
  ADMXRC3_MODEL_TYPE model;
  const char* pName;
} models[] = {
  { ADMXRC3_MODEL_GENERIC, "Generic reconfigurable computing card" },
  { ADMXRC3_MODEL_ADMXRC2, "ADM-XRC-II" },
  { ADMXRC3_MODEL_ADPEXRC5T, "ADPe-XRC-5T" },
  { ADMXRC3_MODEL_ADMXRC6TL, "ADM-XRC-6TL" },
  { ADMXRC3_MODEL_ADMXRC6T1, "ADM-XRC-6T1" }
};

static const char*
getModelName(
  ADMXRC3_MODEL_TYPE modelType)
{
  int i;

  for (i = 0; i < ARRAY_LENGTH(models); i++) {
    if (models[i].model == modelType) {
      return models[i].pName;
    }
  }

  return NULL;
}

static const char*
getSpaceName(
  unsigned int index,
  ADMXRC3_MODEL_TYPE model)
{
  switch (model) {
  case ADMXRC3_MODEL_GENERIC:
    if (index == 0) {
      return "Unknown prefetchable space";
    } else if (index == 1) {
      return "Unknown nonprefetchable space";
    } else if (index == 2) {
      return "Unknown model-specific registers";
    } else if (index == 3) {
      return "ADB3 bridge registers";
    } else {
      return "Unknown";
    }
    break;

  case ADMXRC3_MODEL_ADMXRC2:
    if (index == 0) {
      return "Target FPGA 0 space";
    } else if (index == 1) {
      return "ADM-XRC-II-specific registers";
    } else if (index == 2) {
      return "PCI9656 Registers";
    } else {
      return "Unknown";
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    if (index == 0) {
      return "Target FPGA 0 prefetchable space";
    } else if (index == 1) {
      return "Target FPGA 0 nonprefetchable space";
    } else if (index == 2) {
      return "ADPe-XRC-5T-specific registers";
    } else if (index == 3) {
      return "ADB3 bridge registers";
    } else {
      return "Unknown";
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
    if (index == 0) {
      return "Target FPGA 0 prefetchable space";
    } else if (index == 1) {
      return "Target FPGA 0 nonprefetchable space";
    } else if (index == 2) {
      return "ADM-XRC-6TL-specific registers";
    } else if (index == 3) {
      return "ADB3 bridge registers";
    } else {
      return "Unknown";
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6T1:
    if (index == 0) {
      return "Target FPGA 0 prefetchable space";
    } else if (index == 1) {
      return "Target FPGA 0 nonprefetchable space";
    } else if (index == 2) {
      return "ADM-XRC-6T1-specific registers";
    } else if (index == 3) {
      return "ADB3 bridge registers";
    } else {
      return "Unknown";
    }
    break;

  default:
    return "Unknown";
  }
}

const char*
getUnit(
  ADMXRC3_UNIT_TYPE unit)
{
  switch (unit) {
  case ADMXRC3_UNIT_NONE:
    return "(unitless)";

  case ADMXRC3_UNIT_A:
    return "A";

  case ADMXRC3_UNIT_V:
    return "V";

  case ADMXRC3_UNIT_C:
    return "deg. C";

  case ADMXRC3_UNIT_HZ:
    return "Hz";

  case ADMXRC3_UNIT_RPM:
    return "RPM";

  default:
    return "?";
  }
}

const char*
getDataType(
  ADMXRC3_DATA_TYPE datatype)
{
  switch (datatype) {
  case ADMXRC3_DATA_BOOL:
    return "boolean";

  case ADMXRC3_DATA_DOUBLE:
    return "double";

  case ADMXRC3_DATA_INT32:
    return "int32";

  case ADMXRC3_DATA_UINT32:
    return "uint32";

  default:
    return "?";
  }
}

#define FLAG_SHOWFLASHINFO  (0x10U)  /* Show Flash bank information */
#define FLAG_SHOWMODULEINFO (0x20U)  /* Show I/O module information */
#define FLAG_SHOWSENSORINFO (0x40U)  /* Show sensor information */

STATUS
admxrc3Info(
  unsigned int index,
  unsigned int flags)
{
  STATUS ret = OK;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_BANK_INFO bankInfo;
  ADMXRC3_CARD_INFOEX cardInfoEx;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_MODULE_INFO moduleInfo;
  ADMXRC3_SENSOR_INFO sensorInfo;
  ADMXRC3_VERSION_INFO versionInfo;
  ADMXRC3_WINDOW_INFO windowInfo;
  char modelName[256];
  unsigned int i;
  bool_t bShowFlashInfo, bShowIoModuleInfo, bShowSensorInfo;

  sampleTaskNice();

  bShowFlashInfo = (flags & FLAG_SHOWFLASHINFO) ? TRUE : FALSE;
  bShowIoModuleInfo = (flags & FLAG_SHOWMODULEINFO) ? TRUE : FALSE;
  bShowSensorInfo = (flags & FLAG_SHOWSENSORINFO) ? TRUE : FALSE;

  status = ADMXRC3_Open(index, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  printf("API information\n");
  status = ADMXRC3_GetVersionInfo(hCard, &versionInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get version information: %s\n\n",
      ADMXRC3_GetStatusString(status, TRUE));
  } else {
    printf("API library version           %lu.%lu.%lu\n",
      (unsigned long)versionInfo.Api.Major,
      (unsigned long)versionInfo.Api.Minor,
      (unsigned long)versionInfo.Api.Bugfix);
    printf("Driver version                %lu.%lu.%lu\n",
      (unsigned long)versionInfo.Driver.Major,
      (unsigned long)versionInfo.Driver.Minor,
      (unsigned long)versionInfo.Driver.Bugfix);
    printf("\n");
  }

  status = ADMXRC3_GetCardInfoEx(hCard, &cardInfoEx);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card information: %s\n\n", ADMXRC3_GetStatusString(status, TRUE));
  } else {
    printf("Card information\n");
    if (NULL != getModelName(cardInfoEx.Model)) {
      strcpy(modelName, getModelName(cardInfoEx.Model));
    } else {
      sprintf(modelName, "unknown (0x%lX)", (unsigned long)cardInfoEx.Model);
    }
    printf("Model                         %s\n", modelName);
    printf("Serial number                 %lu(0x%lX)\n", (unsigned long)cardInfoEx.SerialNumber, (unsigned long)cardInfoEx.SerialNumber);
    printf("Number of programmable clocks %lu\n", (unsigned long)cardInfoEx.NumClockGen);
    printf("Number of DMA channels        %lu\n", (unsigned long)cardInfoEx.NumDmaChannel);
    printf("Number of target FPGAs        %lu\n", (unsigned long)cardInfoEx.NumTargetFpga);
    printf("Number of local bus windows   %lu\n", (unsigned long)cardInfoEx.NumWindow);
    printf("Number of sensors             %lu\n", (unsigned long)cardInfoEx.NumSensor);
    printf("Number of I/O module sites    %lu\n", (unsigned long)cardInfoEx.NumModuleSite);
    printf("Number of local bus windows   %lu\n", (unsigned long)cardInfoEx.NumWindow);
    printf("Number of memory banks        %lu\n", (unsigned long)cardInfoEx.NumMemoryBank);
    printf("Bank presence bitmap          0x%lX\n", (unsigned long)cardInfoEx.MemoryBankPresent);
    printf("\n");

    printf("Target FPGA information\n");
    for (i = 0; i < cardInfoEx.NumTargetFpga; i++) {
      printf("FPGA %-2lu                       ", (unsigned long)i);
      status = ADMXRC3_GetFpgaInfo(hCard, i, &fpgaInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        printf("xc%s", fpgaInfo.Identifier);
        if (fpgaInfo.Flags & ADMXRC3_FPGA_SPEEDVALID) {
          printf("-%s", fpgaInfo.SpeedGrade);
        }
        if (fpgaInfo.Flags & ADMXRC3_FPGA_STEPPINGVALID) {
          printf(" step %s", fpgaInfo.Stepping);
        }
        printf("\n");
      }
    }
    if (0 != cardInfoEx.NumTargetFpga) {
      printf("\n");
    } else {
      printf("(none)\n\n");
    }

    printf("Memory bank information\n");
    for (i = 0; i < cardInfoEx.NumMemoryBank; i++) {
      printf("Bank %-2lu                       ", (unsigned long)i);
      status = ADMXRC3_GetBankInfo(hCard, i, &bankInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        if (bankInfo.Present) {
          char unknownTypeStr[40];
          char* pTypeStr;

          if (bankInfo.TypeMask & (ADMXRC3_BANK_ZBT_FT | ADMXRC3_BANK_ZBT_P)) {
            pTypeStr = "SSRAM, ZBT P+F";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_ZBT_FT) {
            pTypeStr = "SSRAM, ZBT F";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_ZBT_P) {
            pTypeStr = "SSRAM, ZBT P";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_SDR) {
            pTypeStr = "SDRAM, SDR";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR) {
            pTypeStr = "SDRAM, DDR";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR2) {
            pTypeStr = "SDRAM, DDR-II";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SSRAM_DDR2) {
            pTypeStr = "SSRAM, DDR-II";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SSRAM_QDR) {
            pTypeStr = "SSRAM, QDR";
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR3) {
            pTypeStr = "SDRAM, DDR3";
          } else {
            sprintf(unknownTypeStr, "Unknown memory type (0x%lX)", (unsigned long)bankInfo.TypeMask);
            pTypeStr = unknownTypeStr;
          }
          printf("%s, %llu kiWord x %lu+%lu bits\n",
            pTypeStr, (unsigned long long)bankInfo.PhysicalSize / 0x400U,
            (unsigned long)bankInfo.PhysicalDataWidth, (unsigned long)bankInfo.PhysicalECCWidth);
          if (0 != bankInfo.MaximumFrequency) {
            printf("                              %.1f MHz - %.1f MHz\n",
              (double)bankInfo.MinimumFrequency / 1000000.0, (double)bankInfo.MaximumFrequency / 1000000.0);
          }
          printf("                              Connectivity mask 0x%lX\n", (unsigned long)bankInfo.ConnectivityMask);
        } else {
          printf("Not present\n");
        }
      }
    }
    if (0 != cardInfoEx.NumMemoryBank) {
      printf("\n");
    } else {
      printf("(none)\n\n");
    }

    printf("Local bus window information\n");
    for (i = 0; i < cardInfoEx.NumWindow; i++) {
      char name[128];

      if (sprintf(name, "(%s)", getSpaceName(i, cardInfoEx.Model)) > 0) {
        printf("Window %-2lu %-.18s  ", (unsigned long)i, name);
      } else {
        printf("Window %-2lu %-.18s  ", (unsigned long)i, "(\?\?\?)");
      }
      status = ADMXRC3_GetWindowInfo(hCard, i, &windowInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        printf("Bus base     0x%llX size 0x%llX\n",
          (unsigned long long)windowInfo.BusBase, (unsigned long long)windowInfo.BusSize);
        printf("                              Local base   0x%llX size 0x%llX\n",
          (unsigned long long)windowInfo.LocalBase, (unsigned long long)windowInfo.LocalSize);
        printf("                              Virtual size 0x%llX\n",
          (unsigned long long)windowInfo.VirtualSize);
      }
    }
  }

  if (bShowFlashInfo) {
    printf("Flash bank information\n");
    for (i = 0; i < cardInfoEx.NumFlashBank; i++) {
      printf("Bank %-2lu                       ", (unsigned long)i);
      status = ADMXRC3_GetFlashInfo(hCard, i, &flashInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        printf("%s, %llu(0x%llX) kiB\n",
          flashInfo.Identifier, (unsigned long long)(flashInfo.Size >> 10), (unsigned long long)(flashInfo.Size >> 10));
        printf("                              Useable area 0x%llX-0x%llX\n",
          (unsigned long long)flashInfo.UseableStart,
          (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1));
      }
    }
    if (0 != cardInfoEx.NumFlashBank) {
      printf("\n");
    } else {
      printf("(none)\n\n");
    }
  }

  if (bShowIoModuleInfo) {
    printf("I/O module information\n");
    for (i = 0; i < cardInfoEx.NumModuleSite; i++) {
      printf("Module %-2lu                     ", (unsigned long)i);
      status = ADMXRC3_GetModuleInfo(hCard, i, &moduleInfo);
      if (status != ADMXRC3_SUCCESS) {
        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        if (moduleInfo.Present) {
          if (moduleInfo.Flags & ADMXRC3_MODULE_LEGACYXRM) {
            printf("Legacy XRM (no further information available)\n");
          } else {
            printf("Product %s\n", moduleInfo.Product);
            printf("                  Part number %s\n", moduleInfo.PartNumber);
            printf("                 Manufacturer %s\n", moduleInfo.Manufacturer);
            printf("                Serial number %s\n", moduleInfo.SerialNumber);
            printf("             Manufacture time %lu minutes since 00:00 1/1/1996\n", (unsigned long) moduleInfo.ManufactureDate);
            printf("                  I/O voltage %.2f\n", moduleInfo.IoVoltage);
          }
          printf("                        Flags 0x%lx%s%s%s%s%s\n",
            (unsigned long)moduleInfo.Flags,
            moduleInfo.Flags ? " => " : "",
            moduleInfo.Flags & ADMXRC3_MODULE_LEGACYXRM ? "LEGACYXRM " : "",
            moduleInfo.Flags & ADMXRC3_MODULE_FORCE2V5 ? "FORCE2V5 " : "",
            moduleInfo.Flags & ADMXRC3_MODULE_ROMERROR ? "ROMERROR " : "",
            moduleInfo.Flags & ADMXRC3_MODULE_CSUMERROR ? "CSUMERROR " : "");
        } else {
          printf("Not present\n");
        }
      }
    }
    if (0 != cardInfoEx.NumModuleSite) {
      printf("\n");
    } else {
      printf("(none)\n\n");
    }
  }

  if (bShowSensorInfo) {
    printf("Sensor information\n");
    for (i = 0; i < cardInfoEx.NumSensor; i++) {
      printf("Sensor %-2lu                     ", (unsigned long)i);
      status = ADMXRC3_GetSensorInfo(hCard, i, &sensorInfo);
      if (status != ADMXRC3_SUCCESS) {

        printf("Failed to get info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      } else {
        printf("%s\n", sensorInfo.Description);
        printf("                              %s, %s, exponent %d, error %.1f\n",
          getUnit(sensorInfo.Unit), getDataType(sensorInfo.DataType), sensorInfo.Exponent, sensorInfo.Error);
        if (sensorInfo.Capabilities) {
          printf("                              %s\n", (sensorInfo.Capabilities & ADMXRC3_SENSOR_SWVALUE) ? "Software sensor" : "");
        }
      }
    }
    if (0 != cardInfoEx.NumSensor) {
      printf("\n");
    } else {
      printf("(none)\n\n");
    }
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
