--==============================================================================
--
--     File Name: adb3_ocp_cross_clk_dom.vhd
--       Version: 0.1
--          Date: 10 Jun 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 OCP cross clock domain
--
--   Description: Asychronous connection between two clock domains for ADB3 OCP
--                transactions
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 10 Jun 2010: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.adb3_ocp.all;

entity adb3_ocp_cross_clk_dom is
port(
  -- Upstream OCP interface
  master_rst    : in    std_logic;
  master_clk    : in    std_logic;
  master_m2s    : in    adb3_ocp_m2sT;
  master_s2m    : out   adb3_ocp_s2mT;
  master_bl_err : out   std_logic;
  -- Downstream OCP interface
  slave_rst     : in    std_logic;
  slave_clk     : in    std_logic;
  slave_m2s     : out   adb3_ocp_m2sT;
  slave_s2m     : in    adb3_ocp_s2mT;
  slave_bl_err  : out   std_logic);
end;

architecture synth of adb3_ocp_cross_clk_dom is

  -- Define FIFO widths
  constant CMD_WIDTH   : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_CMD_WIDTH;
  constant MDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH + ADB3_OCP_DATA_WIDTH/8;
  constant SDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH + ADB3_OCP_RESP_WIDTH + ADB3_OCP_TAG_WIDTH;

  -- cmd FIFO signals
  signal cmd_fifo_data_in  : std_logic_vector(CMD_WIDTH-1 downto 0);
  signal cmd_fifo_data_out : std_logic_vector(CMD_WIDTH-1 downto 0);

  signal cmd_fifo_wadv   : std_logic;
  signal cmd_fifo_radv   : std_logic;
  signal cmd_fifo_empty  : std_logic;
  signal cmd_fifo_nfull  : std_logic;
  signal cmd_fifo_full   : std_logic;

  -- mdata FIFO signals
  signal mdata_fifo_data_in  : std_logic_vector(MDATA_WIDTH-1 downto 0);
  signal mdata_fifo_data_out : std_logic_vector(MDATA_WIDTH-1 downto 0);

  signal mdata_fifo_wadv  : std_logic;
  signal mdata_fifo_radv  : std_logic;
  signal mdata_fifo_empty : std_logic;
  signal mdata_fifo_nfull : std_logic;
  signal mdata_fifo_full  : std_logic;

  -- sdata FIFO signals
  signal sdata_fifo_data_in  : std_logic_vector(SDATA_WIDTH-1 downto 0);
  signal sdata_fifo_data_out : std_logic_vector(SDATA_WIDTH-1 downto 0);

  signal sdata_fifo_wadv  : std_logic;
  signal sdata_fifo_radv  : std_logic;
  signal sdata_fifo_empty : std_logic;
  signal sdata_fifo_nfull : std_logic;
  signal sdata_fifo_full  : std_logic;

  -- FIFO levels used to generate error flags
  signal cmd_rlevel   : std_logic_vector(15 downto 0);
  signal cmd_wlevel   : std_logic_vector(15 downto 0);
  signal mdata_rlevel : std_logic_vector(15 downto 0);
  signal mdata_wlevel : std_logic_vector(15 downto 0);
  signal sdata_rlevel : std_logic_vector(15 downto 0);
  signal sdata_wlevel : std_logic_vector(15 downto 0);

begin

  -- Generate Asychronous FIFOs
  cmd_fifo_i : afifo
    generic map(
      width  => CMD_WIDTH,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => master_rst,
      wk     => master_clk,
      wd     => cmd_fifo_data_in,
      wadv   => cmd_fifo_wadv,
      we     => open,
      wpf    => cmd_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => cmd_wlevel,
      rk     => slave_clk,
      rq     => cmd_fifo_data_out,
      radv   => cmd_fifo_radv,
      rflush => '0',
      re     => cmd_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => cmd_rlevel);

  mdata_fifo_i : afifo
    generic map(
      width  => MDATA_WIDTH,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 2,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => master_rst,
      wk     => master_clk,
      wd     => mdata_fifo_data_in,
      wadv   => mdata_fifo_wadv,
      we     => open,
      wpf    => mdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => mdata_wlevel,
      rk     => slave_clk,
      rq     => mdata_fifo_data_out,
      radv   => mdata_fifo_radv,
      rflush => '0',
      re     => mdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => mdata_rlevel);

  sdata_fifo_i : afifo
    generic map(
      width  => SDATA_WIDTH,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 2,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => slave_rst,
      wk     => slave_clk,
      wd     => sdata_fifo_data_in,
      wadv   => sdata_fifo_wadv,
      we     => open,
      wpf    => sdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => sdata_wlevel,
      rk     => master_clk,
      rq     => sdata_fifo_data_out,
      radv   => sdata_fifo_radv,
      rflush => '0',
      re     => sdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => sdata_rlevel);

  -- Generate error flags if slave data FIFO contains more than 1 element
  -- This is only valid for Direct Slave use when burst length is fixed at 1
  slave_bl_err  <= '1' when (AND_reduce(sdata_wlevel(3 downto 1)) = '0') else '0';
  master_bl_err <= '1' when ( OR_reduce(sdata_rlevel(3 downto 1)) = '1') else '0';


  -- Register Nearly Full Signal
  master_p : process(master_clk)
  begin
    if rising_edge(master_clk) then
      cmd_fifo_full      <= cmd_fifo_nfull;
      cmd_fifo_data_in   <= master_m2s.Tag & master_m2s.BurstLength & master_m2s.Addr & master_m2s.Cmd;

      if ((master_m2s.Cmd = OCP_CMD_WRITE)or(master_m2s.Cmd = OCP_CMD_READ))and(cmd_fifo_full = '0') then
        cmd_fifo_wadv <= '1';
      else
        cmd_fifo_wadv <= '0';
      end if;

      mdata_fifo_full    <= mdata_fifo_nfull;
      mdata_fifo_data_in <= master_m2s.DataByteEn & master_m2s.Data;
      mdata_fifo_wadv    <= master_m2s.DataValid and not mdata_fifo_full;
    end if;
  end process;

  slave_p : process(slave_clk)
  begin
    if rising_edge(slave_clk) then
      sdata_fifo_full    <= sdata_fifo_nfull;
      sdata_fifo_data_in <= slave_s2m.Tag & slave_s2m.Resp & slave_s2m.Data;

      if (slave_s2m.Resp = OCP_RESP_VALID)and(sdata_fifo_full = '0') then
        sdata_fifo_wadv <= '1';
      else
        sdata_fifo_wadv <= '0';
      end if;
    end if;
  end process;


  -- Map Control Signals to FIFOs
  -- cmd_fifo
  master_s2m.CmdAccept  <= not(cmd_fifo_full);
  cmd_fifo_radv         <= (slave_s2m.CmdAccept)and(not(cmd_fifo_empty));

  slave_m2s.Cmd         <= cmd_fifo_data_out(ADB3_OCP_CMD_WIDTH-1 downto 0) when (cmd_fifo_empty = '0') else OCP_CMD_IDLE;
  slave_m2s.Addr        <= cmd_fifo_data_out(ADB3_OCP_ADDR_WIDTH+ADB3_OCP_CMD_WIDTH-1 downto ADB3_OCP_CMD_WIDTH);
  slave_m2s.BurstLength <= cmd_fifo_data_out(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+ADB3_OCP_CMD_WIDTH-1 downto ADB3_OCP_ADDR_WIDTH+ADB3_OCP_CMD_WIDTH);
  slave_m2s.Tag         <= cmd_fifo_data_out(CMD_WIDTH-1 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+ADB3_OCP_CMD_WIDTH);

  -- mdata_fifo
  master_s2m.DataAccept <= not mdata_fifo_full;
  mdata_fifo_radv       <= slave_s2m.DataAccept and not mdata_fifo_empty;

  slave_m2s.Data        <= mdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);
  slave_m2s.DataByteEn  <= mdata_fifo_data_out(MDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  slave_m2s.DataValid   <= not mdata_fifo_empty;

  -- sdata_fifo
  slave_m2s.RespAccept  <= not sdata_fifo_full;
  sdata_fifo_radv       <= master_m2s.RespAccept and not sdata_fifo_empty;

  master_s2m.Data       <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);
  master_s2m.Resp       <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH+1 downto ADB3_OCP_DATA_WIDTH) when sdata_fifo_empty = '0' else OCP_RESP_NONE;
  master_s2m.Tag        <= sdata_fifo_data_out(SDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH+2);

end;