--==============================================================================
--
--     File Name: adb3_ocp_reg_split.vhd
--       Version: 0.1
--          Date: 15 Apr 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: OCP address space demux block (registers)
--
--   Description: Demuxes OCP address space in accordance with addresss range
--                table generic input. Should be connected to low-throughput
--                downstream blocks, eg registers
--
--   Limitations: Does not allow multi-threading of commands
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 15 Apr 2010: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library Work;
use work.adb3_ocp.all;
use work.adb3_target_pkg.all;

entity adb3_ocp_reg_split is
generic (
  addr_range_table : ds_addr_range_table_t);
port(
  -- OCP Interface
  ocp_rst     : in    std_logic;
  ocp_clk     : in    std_logic;
  master_m2s  : in    adb3_ocp_m2sT;
  master_s2m  : out   adb3_ocp_s2mT;
  slaves_m2s  : out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1);
  slaves_s2m  : in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1);
  -- Status interface
  w_addr_oor  : out   std_logic;
  r_addr_oor  : out   std_logic);
end;

architecture mixed of adb3_ocp_reg_split is

  constant NUMBER_OF_SLAVES : natural := addr_range_table'length;

  signal master_s2m_mux   : adb3_ocp_s2mT;
  signal slaves_m2s_int   : adb3_ocp_m2sT_array(0 to NUMBER_OF_SLAVES);
  signal slaves_s2m_int   : adb3_ocp_s2mT_array(0 to NUMBER_OF_SLAVES);
  signal slave_m2s_err    : adb3_ocp_m2sT;
  signal slave_s2m_err    : adb3_ocp_s2mT;

  signal act_cmd_vec      : std_logic_vector(NUMBER_OF_SLAVES downto 0);
  signal act_port_vec     : std_logic_vector(NUMBER_OF_SLAVES downto 0);

  signal err_wr           : std_logic;
  signal err_rd           : std_logic;
  signal err_wr_d1        : std_logic;
  signal err_resp         : std_logic;
  signal err_bl           : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);

  signal busy             : std_logic;
  signal writing          : std_logic;
  signal reading          : std_logic;
  signal data_last        : std_logic;
  signal resp_last        : std_logic;
  signal wburst           : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  signal rburst           : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  signal wburst_completed : std_logic;
  signal rburst_completed : std_logic;

begin

  -- Address range table checking
  -- Check that address ranges do not overlap
  -- Check that mask address is aligned with base address
  gen_addr_range_check1_g: for i in 0 to NUMBER_OF_SLAVES-1 generate
    constant MASK_WIDTH : natural := ds_vec_width(addr_range_table(i).addr_mask);
  begin
    gen_addr_range_check2_g: for j in 0 to NUMBER_OF_SLAVES-1 generate
      constant ADDR_BOT_I : ds_addr_s := addr_range_table(i).addr_base;
      constant ADDR_TOP_I : ds_addr_s := addr_range_table(i).addr_base + addr_range_table(i).addr_mask;
      constant ADDR_BOT_J : ds_addr_s := addr_range_table(j).addr_base;
      constant ADDR_TOP_J : ds_addr_s := addr_range_table(j).addr_base + addr_range_table(j).addr_mask;
    begin
      assert not((((ADDR_BOT_J >= ADDR_BOT_I)and(ADDR_BOT_J <= ADDR_TOP_I))or
                  ((ADDR_TOP_J >= ADDR_BOT_I)and(ADDR_TOP_J <= ADDR_TOP_I)))and(i /= j))
        report "(adb3_ocp_reg_split): Address range entries " & integer'image(j) &
               " and " & integer'image(i) & " overlap in addr_range_table."
        severity failure;
    end generate;

    assert not((OR_reduce(addr_range_table(i).addr_base(MASK_WIDTH-1 downto 0))) = '1')
      report "(adb3_ocp_reg_split): Mask/Base address alignment problem for entry " & integer'image(i) & " in addr_range_table."
      severity failure;
  end generate;


  gen1_slave_m2s_ports_g : for i in 0 to NUMBER_OF_SLAVES generate
    -- Generate slaves_m2s OCP ports from master_m2s
    -- Only pass on command if we are not in a transaction
    -- Only validate data if we are in a write command
    -- Only accept dsta if we are in a read command
    slaves_m2s_int(i).Cmd         <= (master_m2s.Cmd)and(not(busy & busy & busy)) when act_cmd_vec(i)  = '1' else OCP_CMD_IDLE;
    slaves_m2s_int(i).DataValid   <= (master_m2s.DataValid)and(writing)           when act_port_vec(i) = '1' else '0';
    slaves_m2s_int(i).RespAccept  <= (master_m2s.RespAccept)and(reading)          when act_port_vec(i) = '1' else '0';
    slaves_m2s_int(i).Addr        <= master_m2s.Addr;
    slaves_m2s_int(i).BurstLength <= master_m2s.BurstLength;
    slaves_m2s_int(i).Data        <= master_m2s.Data;
    slaves_m2s_int(i).DataByteEn  <= master_m2s.DataByteEn;
    slaves_m2s_int(i).Tag         <= master_m2s.Tag;
  end generate;

  slaves_m2s    <= slaves_m2s_int(0 to NUMBER_OF_SLAVES-1);
  slave_m2s_err <= slaves_m2s_int(NUMBER_OF_SLAVES);


  -- Test address ranges to generate active command vector
  act_cmd_vec_p : process(master_m2s)
    variable mask_width  : natural;
    variable act_cmd_var : std_logic_vector(NUMBER_OF_SLAVES-1 downto 0);
  begin
    act_cmd_vec_l : for n in 0 to NUMBER_OF_SLAVES-1 loop
      mask_width := ds_vec_width(addr_range_table(n).addr_mask);

      if (master_m2s.Addr(DS_ADDR_WIDTH-1 downto mask_width) = addr_range_table(n).addr_base(DS_ADDR_WIDTH-1 downto mask_width)) then
        act_cmd_var(n) := '1';
      else
        act_cmd_var(n) := '0';
      end if;
    end loop;

    -- Test for invalid address and set MSB of vectors
    if (OR_reduce(act_cmd_var) = '1') then
      act_cmd_vec <= '0' & act_cmd_var;
    else
      act_cmd_vec <= '1' & act_cmd_var;
    end if;
  end process;


  -- Report if more that one bit of act_cmd_vec is active
  act_cmd_vec_check_p : process(act_cmd_vec)
    variable act_cmd_cnt : natural;
  begin
    act_cmd_cnt := 0;
    gen_act_cmd_check_l: for i in 0 to NUMBER_OF_SLAVES loop
      if (act_cmd_vec(i) = '1') then
        act_cmd_cnt := act_cmd_cnt + 1;
      end if;
    end loop;
    assert not(act_cmd_cnt > 1)
      report "(adb3_ocp_reg_split): Multiple address ranges decoded as active in act_cmd_vec"
      severity failure;
  end process;


  -- Retime act_cmd_vec on CmdAccept to generate port range vector
  act_port_vec_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      act_port_vec(NUMBER_OF_SLAVES)            <= '1';
      act_port_vec(NUMBER_OF_SLAVES-1 downto 0) <= (others => '0');
      writing          <= '0';
      reading          <= '0';
      w_addr_oor       <= '0';
      r_addr_oor       <= '0';
      wburst           <= (others => '0');
      rburst           <= (others => '0');
      wburst_completed <= '0';
      rburst_completed <= '0';
    elsif rising_edge(ocp_clk) then
      -- Timing of act_port_vec is critical for OCP-OCP simulation.
      -- Must be delayed by only 1 cycle from act_cmd_vec
      if (master_m2s.Cmd /= OCP_CMD_IDLE)and(master_s2m_mux.CmdAccept = '1') then
        act_port_vec <= act_cmd_vec;

        if (OR_reduce(act_cmd_vec(NUMBER_OF_SLAVES-1 downto 0)) = '0') then
          if (master_m2s.Cmd = OCP_CMD_WRITE) then
            w_addr_oor <= '1';
          else
            r_addr_oor <= '1';
          end if;
        end if;
      else
        w_addr_oor <= '0';
        r_addr_oor <= '0';
      end if;

      -- Master OCP interface Write transaction
      if (master_m2s.Cmd = OCP_CMD_WRITE)and(master_s2m_mux.CmdAccept = '1')and(busy = '0') then
        writing <= '1';
      elsif ((data_last = '1')and(master_m2s.DataValid = '1')and(master_s2m_mux.DataAccept = '1'))or(wburst_completed = '1') then
        writing <= '0';
      end if;

      -- Generate burst counter wburst and wburst_completed
      wburst_completed <= '0';
      if (master_m2s.Cmd = OCP_CMD_WRITE)and(master_s2m_mux.CmdAccept = '1') then
        if (master_m2s.DataValid = '1')and(master_s2m_mux.DataAccept = '1') then
          wburst <= master_m2s.BurstLength-1;
          if (master_m2s.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            wburst_completed <= '1';
          end if;
        else
          wburst <= master_m2s.BurstLength;
        end if;
      else
        if (master_m2s.DataValid = '1')and(master_s2m_mux.DataAccept = '1') then
          wburst <= wburst-1;
        end if;
      end if;

      -- Master OCP interface Read transaction
      if (master_m2s.Cmd = OCP_CMD_READ)and(master_s2m_mux.CmdAccept = '1')and(busy = '0') then
        reading <= '1';
      elsif ((resp_last = '1')and(master_m2s.RespAccept = '1')and(master_s2m_mux.Resp = OCP_RESP_VALID))or(rburst_completed = '1') then
        reading <= '0';
      end if;

      -- Generate burst counter rburst and rburst_completed
      rburst_completed <= '0';
      if (master_m2s.Cmd = OCP_CMD_READ)and(master_s2m_mux.CmdAccept = '1') then
        if (master_m2s.RespAccept = '1')and(master_s2m_mux.Resp = OCP_RESP_VALID)then
          rburst <= master_m2s.BurstLength-1;
          if (master_m2s.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            rburst_completed <= '1';
          end if;
        else
          rburst <= master_m2s.BurstLength;
        end if;
      else
        if (master_m2s.RespAccept = '1')and(master_s2m_mux.Resp = OCP_RESP_VALID) then
          rburst <= rburst-1;
        end if;
      end if;
    end if;
  end process;

  busy <= writing or reading;

  data_last <= '1' when wburst = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH)) else '0';
  resp_last <= '1' when rburst = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH)) else '0';


  -- Connect Error OCP port for out of range transactions
  slave_s2m_err_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      err_bl        <= (others => '0');
      err_wr        <= '0';
      err_rd        <= '0';
      err_wr_d1     <= '0';
      err_resp      <= '0';
      slave_s2m_err <= ADB3_OCP_DEFAULT_S2M;
    elsif rising_edge(ocp_clk) then
      -- Command Accept
      if ((slave_m2s_err.Cmd = OCP_CMD_WRITE)or(slave_m2s_err.Cmd = OCP_CMD_READ))and(slave_s2m_err.CmdAccept = '0') then
        slave_s2m_err.CmdAccept <= '1';
        if (slave_m2s_err.Cmd = OCP_CMD_WRITE) then
          err_bl <= slave_m2s_err.BurstLength;
          err_wr <= '1';
          err_rd <= '0';
        end if;
        if (slave_m2s_err.Cmd = OCP_CMD_READ) then
          slave_s2m_err.Data <= X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE";
          slave_s2m_err.Tag  <= slave_m2s_err.Tag;
          err_bl <= slave_m2s_err.BurstLength;
          err_rd <= '1';
          err_wr <= '0';
        end if;
      else
        slave_s2m_err.CmdAccept <= '0';
      end if;

      -- Data Accept
      if (err_wr = '1')and(slave_s2m_err.CmdAccept = '1')  then
        err_wr_d1 <= '1';
      end if;

      if (err_wr_d1 = '1')and(slave_m2s_err.DataValid = '1') then
        err_bl <= err_bl - 1;
        slave_s2m_err.DataAccept <= '1';
        if (err_bl = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
          err_wr_d1 <= '0';
        end if;
      else
        slave_s2m_err.DataAccept <= '0';
      end if;

      -- Resp
      if (err_rd = '1')and(slave_s2m_err.CmdAccept = '1')  then
        err_resp <= '1';
      end if;

      if (err_resp = '1')and(slave_m2s_err.RespAccept = '1') then
        err_bl <= err_bl - 1;
        if (err_bl = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
          err_resp <= '0';
        end if;
      end if;

      if (err_resp = '1') then
        slave_s2m_err.Resp <= OCP_RESP_VALID;
      else
        slave_s2m_err.Resp <= OCP_RESP_NONE;
      end if;
    end if;
  end process;


  -- Connect slaves_s2m and slave_s2m_err to slaves_s2m_int
  slaves_s2m_int(0 to NUMBER_OF_SLAVES-1) <= slaves_s2m;
  slaves_s2m_int(NUMBER_OF_SLAVES)        <= slave_s2m_err;

  gen_s2m_ports: for i in 0 to NUMBER_OF_SLAVES generate
    -- Connect slave and error s2m OCP ports to master
    -- Only accept command if we have finished the previous one
    -- Only accept data if we are in a write command
    -- Only respond with data if we are in a read command
    master_s2m_mux.CmdAccept  <= (slaves_s2m_int(i).CmdAccept)and(not(busy))    when act_cmd_vec(i)  = '1' else 'Z';
    master_s2m_mux.DataAccept <= (slaves_s2m_int(i).DataAccept)and(writing)     when act_port_vec(i) = '1' else 'Z';
    master_s2m_mux.Resp       <= (slaves_s2m_int(i).Resp)and(reading & reading) when act_port_vec(i) = '1' else (others => 'Z');
    master_s2m_mux.Data       <= slaves_s2m_int(i).Data                         when act_port_vec(i) = '1' else (others => 'Z');
    master_s2m_mux.Tag        <= slaves_s2m_int(i).Tag                          when act_port_vec(i) = '1' else (others => 'Z');
  end generate;

  master_s2m <= master_s2m_mux;

end;