--==============================================================================
--
--     File Name: cdc_pkg.vhd
--       Version: 0.1
--          Date: 03 Dec 2009
--         Model: RTL (SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Clock domain crossing package
--
--   Description: Clock domain crossing package and entities
--                  toggle_pulse
--                  toggle_pulse_data
--                  send_ack_data
--
--   Limitations:
--
--         Notes: See "Clock Domain Crossing (CDC) Design & Verification
--                Techniques Using System Verilog" Clifford E. Cummings
--                http://www.sunburst-design.com/papers/
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 03 Dec 09: Initial design
--
--
--==============================================================================
library ieee;
use ieee.std_logic_1164.all;

package cdc_pkg is

  ----------------------------
  -- Basic block components --
  ----------------------------

  component syncn
  generic(
    n : natural := 2);
  port(
    clk     : in   std_logic;
    d       : in   std_logic;
    q       : out  std_logic);
  end component;

  component pgen
  port(
    clk     : in   std_logic;
    toggle  : in   std_logic;
    q       : out  std_logic;
    pulse   : out  std_logic);
  end component;

  component pgen_data
  generic(
    width : natural := 8);
  port(
    clk     : in   std_logic;
    data_i  : in   std_logic_vector(width-1 downto 0);
    toggle  : in   std_logic;
    data_o  : out  std_logic_vector(width-1 downto 0);
    valid   : out  std_logic);
  end component;

  component toggle
  port(
    clk    : in   std_logic;
    pulse  : in   std_logic;
    toggle : out  std_logic);
  end component;

  component toggle_data
  generic(
    width : natural := 8);
  port(
    clk      : in   std_logic;
    data_i   : in   std_logic_vector(width-1 downto 0);
    pulse    : in   std_logic;
    toggle   : out  std_logic;
    data_o   : out  std_logic_vector(width-1 downto 0));
  end component;

  component send_fsm is
  port(
    clk    : in   std_logic;
    send   : in   std_logic;
    ack    : in   std_logic;
    ready  : out  std_logic);
  end component;

  component ack_fsm is
  port(
    clk    : in   std_logic;
    load   : in   std_logic;
    en     : in   std_logic;
    valid  : out  std_logic);
  end component;

  component send_data is
  generic(
    width : natural := 8);
  port(
    clk    : in   std_logic;
    data_i : in   std_logic_vector(width-1 downto 0);
    send   : in   std_logic;
    ready  : out  std_logic;
    data_o : out  std_logic_vector(width-1 downto 0);
    ack    : in   std_logic;
    en     : out  std_logic);
  end component;

  component ack_data
  generic(
    width : natural := 8);
  port(
    clk    : in   std_logic;
    data_i : in   std_logic_vector(width-1 downto 0);
    load   : in   std_logic;
    valid  : out  std_logic;
    data_o : out  std_logic_vector(width-1 downto 0);
    en     : in   std_logic;
    ack    : out  std_logic);
  end component;


  -----------------------------------
  -- Cross clock domain components --
  -----------------------------------

  component toggle_pulse
  port(
    clk_a   : in   std_logic;
    in_a    : in   std_logic;
    clk_b   : in   std_logic;
    out_b   : out  std_logic);
  end component;

  component toggle_pulse_data
  generic(
    width : natural := 8);
  port(
    clk_a    : in   std_logic;
    send_a   : in   std_logic;
    data_a   : in   std_logic_vector(width-1 downto 0);
    clk_b    : in   std_logic;
    data_b   : out  std_logic_vector(width-1 downto 0);
    valid_b  : out  std_logic);
  end component;

  component send_ack_data
  generic(
    width : natural := 8);
  port(
    clk_a    : in   std_logic;
    send_a   : in   std_logic;
    data_a   : in   std_logic_vector(width-1 downto 0);
    ready_a  : out  std_logic;
    clk_b    : in   std_logic;
    load_b   : in   std_logic;
    data_b   : out  std_logic_vector(width-1 downto 0);
    valid_b  : out  std_logic);
  end component;

end package cdc_pkg;


package body cdc_pkg is

end package body;

-----------------------------------------------------
-- Basic blocks - entirely within one clock domain --
-----------------------------------------------------

--
-- syncn
-- Synchronise signal d to different clock domain clk
-- Select number of re-times for metastability using n
--
library ieee;
use ieee.std_logic_1164.all;

entity syncn is
generic(
  n : natural := 2);
port(
  clk     : in   std_logic;
  d       : in   std_logic;
  q       : out  std_logic);
end entity syncn;

architecture syn of syncn is
  type pipe_array_t is array(0 to n-1) of std_logic;
  signal pipe_array : pipe_array_t := (others => '0');
begin
  syncn_p : process(clk)
  begin
    if rising_edge(clk) then
      pipe_array <= d & pipe_array(0 to n-2);
    end if;
  end process;

  q <= pipe_array(n-1);
end architecture;


--
-- pgen
-- Generate 1-cycle pulse from edge on syncronised signal
-- Synchronised signal produced by syncn
--
library ieee;
use ieee.std_logic_1164.all;

entity pgen is
port(
  clk     : in   std_logic;
  toggle  : in   std_logic;
  q       : out  std_logic;
  pulse   : out  std_logic);
end entity pgen;

architecture syn of pgen is
  signal pipe_q  : std_logic := '0';
begin
  -- further re-time for pulse generation
  retime_p : process(clk)
  begin
    if rising_edge(clk) then
      pipe_q <= toggle;
    end if;
  end process;

  -- output q synchronised to clk clock domain when used in
  -- conjuction with syncn blobk (n+1 cycle delay)
  -- output edge detect generated pulse
  q     <= pipe_q;
  pulse <= toggle xor pipe_q;
end architecture;


--
-- pgen_data
-- Generate 1-cycle pulse from edge on syncronised signal
-- Synchronised signal produced by syncn
-- Use generated pulse to load data of width width
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity pgen_data is
generic(
  width : natural := 8);
port(
  clk     : in   std_logic;
  data_i  : in   std_logic_vector(width-1 downto 0);
  toggle  : in   std_logic;
  data_o  : out  std_logic_vector(width-1 downto 0);
  valid   : out  std_logic);
end entity pgen_data;

architecture syn of pgen_data is
  signal pulse     : std_logic;
  signal data_ret  : std_logic_vector(width-1 downto 0) := (others => '0');
  signal valid_int : std_logic := '0';
begin
  -- generate pulse from toggle
  pgen_i : pgen
  port map(
    clk     => clk,
    toggle  => toggle,
    q       => open,
    pulse   => pulse);

  -- Sample input data_i on pulse
  ret_data_p : process(clk)
  begin
    if rising_edge(clk) then
      if (pulse = '1') then
        data_ret <= data_i;
      end if;
      valid_int <= pulse;
    end if;
  end process;

  data_o <= data_ret;
  valid  <= valid_int;
end architecture;


--
-- toggle
-- Generate a toggling signal from an change in level on input pulse
--
library ieee;
use ieee.std_logic_1164.all;

entity toggle is
port(
  clk    : in   std_logic;
  pulse  : in   std_logic;
  toggle : out  std_logic);
end entity toggle;

architecture syn of toggle is
  signal toggle_int : std_logic := '0';
begin
  -- generate toggle signal from input pulse
  retime_p : process(clk)
  begin
    if rising_edge(clk) then
      toggle_int <= toggle_int xor pulse;
    end if;
  end process;
  toggle <= toggle_int;
end architecture;


--
-- toggle_data
-- Generate a toggling signal from an change in level on input pulse
-- Use input pulse to load data of width width
--
library ieee;
use ieee.std_logic_1164.all;

entity toggle_data is
generic(
  width : natural := 8);
port(
  clk    : in   std_logic;
  data_i : in   std_logic_vector(width-1 downto 0);
  pulse  : in   std_logic;
  toggle : out  std_logic;
  data_o : out  std_logic_vector(width-1 downto 0));
end entity toggle_data;

architecture syn of toggle_data is
  signal data_ret   : std_logic_vector(width-1 downto 0) := (others => '0');
  signal toggle_int : std_logic := '0';
begin
  -- Sample input data_i on pulse
  ret_data_p : process(clk)
  begin
    if rising_edge(clk) then
      if (pulse = '1') then
        data_ret <= data_i;
      end if;
    end if;
  end process;

  -- generate toggle signal from pulse
  gen_toggle_p : process(clk)
  begin
    if rising_edge(clk) then
      toggle_int <= toggle_int xor pulse;
    end if;
  end process;

  toggle <= toggle_int;
  data_o <= data_ret;
end architecture;


--
-- send_fsm
-- FSM for sending side of closed-loop multi-cycle path
-- send + acknowledge
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity send_fsm is
port(
  clk    : in   std_logic;
  send   : in   std_logic;  -- Send data
  ack    : in   std_logic;  -- Acknowledge receipt of data
  ready  : out  std_logic); -- Ready to send next data
end entity send_fsm;

architecture syn of send_fsm is
  type state_t is (READY_ST,BUSY_ST);
  signal curr_st : state_t := READY_ST;
  signal next_st : state_t := READY_ST;
begin

  update_state_p : process(clk)
  begin
    if rising_edge(clk) then
      curr_st <= next_st;
    end if;
  end process;

  gen_next_state_p : process(curr_st,send,ack)
  begin
    case curr_st is
      when READY_ST =>
        if (send = '1') then
          next_st <= BUSY_ST;
        else
          next_st <= READY_ST;
        end if;
        ready <= '1';

      when BUSY_ST =>
        if (ack = '1') then
          next_st <= READY_ST;
        else
          next_st <= BUSY_ST;
        end if;
        ready <= '0';
    end case;
  end process;
end architecture;


--
-- ack_fsm
-- FSM for acknowledge side of closed-loop multi-cycle path
-- send + acknowledge
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity ack_fsm is
port(
  clk    : in   std_logic;
  load   : in   std_logic;  -- Load data / send acknowledge
  en     : in   std_logic;  -- Enable receipt of data
  valid  : out  std_logic); -- Data valid / ready to load
end entity ack_fsm;

architecture syn of ack_fsm is
  type state_t is (READY_ST,WAIT_ST);
  signal curr_st : state_t := READY_ST;
  signal next_st : state_t := READY_ST;
begin

  update_state_p : process(clk)
  begin
    if rising_edge(clk) then
      curr_st <= next_st;
    end if;
  end process;

  gen_next_state_p : process(curr_st,load,en)
  begin
    case curr_st is
      when READY_ST =>
        if (load = '1') then
          next_st <= WAIT_ST;
        else
          next_st <= READY_ST;
        end if;
        valid <= '1';

      when WAIT_ST =>
        if (en = '1') then
          next_st <= READY_ST;
        else
          next_st <= WAIT_ST;
        end if;
        valid <= '0';
    end case;
  end process;
end architecture;


--
-- send_data
-- Send a 1-cycle enable pulse between clock domains.
-- Receive a 1-cycle acknowledge pulse between clock domains.
-- enable is generated from send signal and send FSM output ready
-- acknowledge is generated by ack_data block
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity send_data is
generic(
  width : natural := 8);
port(
  clk    : in   std_logic;
  data_i : in   std_logic_vector(width-1 downto 0);
  send   : in   std_logic;
  ready  : out  std_logic;
  data_o : out  std_logic_vector(width-1 downto 0);
  ack    : in   std_logic;
  en     : out  std_logic);
end entity send_data;

architecture syn of send_data is
  signal ld        : std_logic;
  signal ack_int   : std_logic;
  signal data_ret  : std_logic_vector(width-1 downto 0) := (others => '0');
  signal ready_int : std_logic := '0';
  signal en_int    : std_logic := '0';
begin
  ld <= send and ready_int;

  -- generate en_int and sample data_i on ld pulse
  toggle_data_i : toggle_data
  generic map(
    width => width)
  port map(
    clk    => clk,
    data_i => data_i,
    pulse  => ld,
    toggle => en_int,
    data_o => data_ret);

  -- generate pulse from ack
  pgen_i : pgen
  port map(
    clk     => clk,
    toggle  => ack,
    q       => open,
    pulse   => ack_int);

  -- send fsm: generate ready from send and ack
  send_fsm_i : send_fsm
  port map(
    clk     => clk,        -- in
    send    => send,       -- in
    ack     => ack_int,    -- in
    ready   => ready_int); -- out

  data_o <= data_ret;
  ready  <= ready_int;
  en     <= en_int;
end architecture;


--
-- ack_data
-- Send a 1-cycle acknowledge pulse between clock domains.
-- Receive a 1-cycle enable pulse between clock domains.
-- acknowledge is generated from load signal and send FSM output valid
-- enable is generated by send_data block
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity ack_data is
generic(
  width : natural := 8);
port(
  clk    : in   std_logic;
  data_i : in   std_logic_vector(width-1 downto 0);
  load   : in   std_logic;
  valid  : out  std_logic;
  data_o : out  std_logic_vector(width-1 downto 0);
  en     : in   std_logic;
  ack    : out  std_logic);
end entity ack_data;

architecture syn of ack_data is
  signal ld        : std_logic;
  signal en_int    : std_logic;
  signal data_ret  : std_logic_vector(width-1 downto 0) := (others => '0');
  signal valid_int : std_logic := '0';
  signal ack_int   : std_logic := '0';
begin
  ld <= valid_int and load;

  -- generate ack and retime data_i
  toggle_data_i : toggle_data
  generic map(
    width => width)
  port map(
    clk    => clk,
    data_i => data_i,
    pulse  => ld,
    toggle => ack_int,
    data_o => data_ret);

  -- generate en pulse from en toggle input
  pgen_i : pgen
  port map(
    clk     => clk,
    toggle  => en,
    q       => open,
    pulse   => en_int);

  -- acknowledge fsm: generate valid from en and load
  ack_fsm_i : ack_fsm
  port map(
    clk    => clk,         -- in
    load   => load,        -- in
    en     => en_int,      -- in
    valid  => valid_int);  -- out

  data_o <= data_ret;
  valid  <= valid_int;
  ack    <= ack_int;
end architecture;


----------------------------------------------------------------------------
-- Cross clock domain solution blocks instantiating basic blocks only     --
-- Each basic block is in one clock domain only. Signals are synchronised --
-- between clock domains using the syncn basic block                      --
----------------------------------------------------------------------------

--
-- toggle_pulse
-- Open-loop multi-cycle path toggle-pulse
-- Independent of clka/clkb frequency relationship
-- Instantiation of blocks in different clock domains only
--
-- Pass a 1-cycle control signal pulse between clock domains.
-- Use the control signal to generate a toggling signal in the
-- source clock domain that is synchronised and passed to a pulse
-- generator in the destination clock domain.
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity toggle_pulse is
port(
  clk_a    : in   std_logic;   -- clk_a domain:
  in_a     : in   std_logic;   -- clk_a domain: input 1-cycle pulse
  clk_b    : in   std_logic;   -- clk_b domain:
  out_b    : out  std_logic);  -- clk_b domain: output 1-cycle pulse
end entity toggle_pulse;

architecture syn of toggle_pulse is
  -- clk_a domain signals
  signal toggle_a : std_logic;
  -- clk_b domain signals
  signal toggle_b : std_logic;
begin
  -- generate toggle_a signal from input pulse in clk_a domain
  toggle_i : toggle
  port map(
    clk    => clk_a,
    pulse  => in_a,
    toggle => toggle_a);

  -- Synchronise toggle_a into clk_b domain
  sync2_i : syncn
  port map(
    clk   => clk_b,
    d     => toggle_a,
    q     => toggle_b);

  -- generate pulse from toggle_b signal in clk_b domain
  pgen_i : pgen
  port map(
    clk    => clk_b,
    toggle => toggle_b,
    q      => open,
    pulse  => out_b);
end architecture;


--
-- toggle_pulse_data
-- Open-loop multi-cycle path toggle-pulse + data
-- Independent of clka/clkb frequency relationship
-- Instantiation of blocks in different clock domains only
--
-- Pass a 1-cycle control signal pulse between clock domains.
-- Use the control signal to generate a toggling signal in the
-- source clock domain that is synchronised and passed to a pulse
-- generator in the destination clock domain.
-- Use control signals to safely pass data between clock domains
--
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity toggle_pulse_data is
generic(
  width : natural := 8);
port(
  clk_a    : in   std_logic;                            -- clk_a domain:
  send_a   : in   std_logic;                            -- clk_a domain: send data valid at input
  data_a   : in   std_logic_vector(width-1 downto 0);   -- clk_a domain: input data
  clk_b    : in   std_logic;                            -- clk_b domain:
  data_b   : out  std_logic_vector(width-1 downto 0);   -- clk_b domain: output data
  valid_b  : out  std_logic);                           -- clk_b domain: receive data valid at output
end entity toggle_pulse_data;

architecture syn of toggle_pulse_data is
  signal data_tx  : std_logic_vector(width-1 downto 0);
  -- clk_a domain signals
  signal toggle_a : std_logic;
  -- clk_b domain signals
  signal toggle_b : std_logic;
begin
  -- generate toggle and sample data_a in clk_a domain
  toggle_data_i : toggle_data
  generic map(
    width => width)
  port map(
    clk    => clk_a,
    data_i => data_a,
    pulse  => send_a,
    toggle => toggle_a,
    data_o => data_tx);

  -- synchronise toggle_a into clk_b domain
  sync2_i : syncn
  port map(
    clk   => clk_b,
    d     => toggle_a,
    q     => toggle_b);

  -- generate pulse from toggle_b signal in clk_b domain
  pgen_data_i : pgen_data
  generic map(
    width => width)
  port map(
    clk     => clk_b,
    data_i  => data_tx,
    toggle  => toggle_b,
    data_o  => data_b,
    valid   => valid_b);
end architecture;


--
-- send_ack_data
-- Closed-loop multi-cycle path send + acknowledge
-- Independent of clka/clkb frequency relationship
-- Instantiation of blocks in different clock domains only
library ieee;
use ieee.std_logic_1164.all;

library work;
use work.cdc_pkg.all;

entity send_ack_data is
generic(
  width : natural := 8);
port(
  clk_a    : in   std_logic;                            -- clk_a domain:
  send_a   : in   std_logic;                            -- clk_a domain: send valid input data
  data_a   : in   std_logic_vector(width-1 downto 0);   -- clk_a domain: input data
  ready_a  : out  std_logic;                            -- clk_a domain: ready to send next input data
  clk_b    : in   std_logic;                            -- clk_b domain:
  load_b   : in   std_logic;                            -- clk_b domain: ready to receive next input data
  data_b   : out  std_logic_vector(width-1 downto 0);   -- clk_b domain: output data
  valid_b  : out  std_logic);                            -- clk_b domain: receive valid output data
end entity send_ack_data;

architecture syn of send_ack_data is
  signal data_tx  : std_logic_vector(width-1 downto 0);
  -- clk_a domain signals
  signal en_a       : std_logic;
  signal ack_a      : std_logic;
  -- clk_b domain signals
  signal en_b       : std_logic;
  signal ack_b      : std_logic;
begin
  -- generate en and sample data_a in clk_a domain
  send_data_i : send_data
  generic map(
    width => width)
  port map(
    clk    => clk_a,
    data_i => data_a,
    send   => send_a,
    ready  => ready_a,
    data_o => data_tx,
    ack    => ack_a,
    en     => en_a);

  -- synchronise en_a into clk_b domain
  sync2_en_i : syncn
  port map(
    clk   => clk_b,
    d     => en_a,
    q     => en_b);

  -- synchronise ack_b into clk_a domain
  sync2_ack_i : syncn
  port map(
    clk   => clk_a,
    d     => ack_b,
    q     => ack_a);

  -- generate ack and generate data_b in clk_b domain
  ack_data_i : ack_data
  generic map(
    width => width)
  port map(
    clk     => clk_b,
    data_i  => data_tx,
    load    => load_b,
    valid   => valid_b,
    data_o  => data_b,
    en      => en_b,
    ack     => ack_b);
end architecture;
