--==============================================================================
--
--     File Name: clock_speed_pkg.vhd
--       Version: 0.4
--          Date: 18 May 2010
--         Model: RTL (SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Clock Frequency measurement
--
--   Description: Measures the number of edges on smp_clk, during a defined
--                measurement period (cycles of ref_clk), to produce the
--                frequency of smp_clk. Different meaurement periods will
--                produce results with accuracies as follows:
--                  (ref_clk freq / 1 MHz) cycles => smp_clk freq in MHz
--                  (ref_clk freq / 1 KHz) cycles => smp_clk freq in KHz
--                  (ref_clk freq / 1  Hz) cycles => smp_clk freq in  Hz
--
--   Limitations: As synchronisation of signals from ref_clk to smp_clk domain
--                requires 2 smp_clk edges, measurement of smp_clk frequencies
--                > 0 but < 2 Hz/KHz/MHz will be unreliable.
--                Measurement of smp_clk frequencies is limited by size of
--                smp_cnt register. 29 bits => 536.870911 MHz.
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, David Dolman, 10 Jul 09: Initial design
--                0.2, KDC         , 04 Dec 09: Changes:
--                        Replaced the cross-clock domain signal handling to
--                        use the cdc package.
--                        Reduced the measurement options to Hz, KHz, MHz
--                        Added 1 to smp_cnt_last value
--                0.3, KDC         , 27 Jan 10: Changes:
--                        Added generic for ref_clk frequency in MHz
--                        Changed generic for accuracy selection
--                        Moved ref_clk_cnt_max calculation from pkg to entity
--                        Changed ref_clk_cnt from up counter to down counter
--                        Added subtype for smp_clk_cnt
--                0.4, KDC         , 18 May 10: Changes:
--                        Added use of read_running to select between generation
--                        of freq/valid in smp_clk domain and freq/valid in
--                        ref_clk domain
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

package clock_speed_pkg is

  type freq_meas_type_t is (FREQ_IN_HZ,FREQ_IN_KHZ,FREQ_IN_MHZ);

  subtype ref_clk_cnt_s is std_logic_vector(27 downto 0);
  subtype smp_clk_cnt_s is std_logic_vector(31 downto 0);

  component clock_speed
  generic(
    ref_clk_freq_hz : natural := 200000000;
    freq_meas_type  : freq_meas_type_t := FREQ_IN_HZ);
  port(
    ref_clk         : in  std_logic;
    smp_clk         : in  std_logic;
    read_clk        : in  std_logic;
    read_freq       : out std_logic_vector(31 downto 0);
    read_update     : out std_logic;
    read_running    : out std_logic);
  end component;

end package clock_speed_pkg;


package body clock_speed_pkg is

end package body;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library adcommon;
use adcommon.cdc_pkg.toggle_pulse;
use adcommon.cdc_pkg.toggle_pulse_data;
use adcommon.cdc_pkg.syncn;
use adcommon.clock_speed_pkg.all;

entity clock_speed is
generic(
  ref_clk_freq_hz : natural := 200000000;
  freq_meas_type  : freq_meas_type_t := FREQ_IN_HZ);
port(
  ref_clk         : in  std_logic;
  smp_clk         : in  std_logic;
  read_clk        : in  std_logic;
  read_freq       : out std_logic_vector(31 downto 0);
  read_update     : out std_logic;
  read_running    : out std_logic);
end entity clock_speed;


architecture syn of clock_speed is

  -- eg. for 200 MHz:
  -- 200000000/1       * 5 ns = 1.000000s (1s)  sample period => result in Hz*1       (Hz)
  -- 200000000/1000    * 5 ns = 0.001000s (1ms) sample period => result in Hz*1000    (KHz)
  -- 200000000/1000000 * 5 ns = 0.000001s (1us) sample period => result in Hz*1000000 (MHz)
  constant REF_CLK_CNT_MAX_HZ  : ref_clk_cnt_s := std_logic_vector(to_unsigned(ref_clk_freq_hz,28));
  constant REF_CLK_CNT_MAX_KHZ : ref_clk_cnt_s := std_logic_vector(to_unsigned(ref_clk_freq_hz/1000,28));
  constant REF_CLK_CNT_MAX_MHZ : ref_clk_cnt_s := std_logic_vector(to_unsigned(ref_clk_freq_hz/1000000,28));

  constant REF_CLK_CNT_8 : ref_clk_cnt_s := x"0000008";
  constant REF_CLK_CNT_2 : ref_clk_cnt_s := x"0000002";

  constant SMP_CNT_0     : smp_clk_cnt_s := x"00000000";
  constant SMP_CNT_1     : smp_clk_cnt_s := x"00000001";

  -- ref_clk domain signals
  -- Initisalise ref_clk_cnt to just before tc value so that we produce an initial
  -- ref_clk_tc to initialise the smp_cnt counter
  signal ref_clk_cnt         : ref_clk_cnt_s := REF_CLK_CNT_8;
  signal ref_clk_cnt_max     : ref_clk_cnt_s;
  signal ref_clk_tc          : std_logic := '0';
  signal smp_clk_active      : std_logic := '0';
  signal smp_clk_edge        : std_logic := '0';

  -- smp_clk domain signals
  signal ref_clk_tc_smp_clk : std_logic;
  signal smp_cnt            : smp_clk_cnt_s := (others => '0');
  signal smp_cnt_last       : smp_clk_cnt_s := (others => '0');
  signal smp_valid          : std_logic := '0';
  signal smp_valid_ref_clk  : std_logic;

  -- read_clk domain signals
  signal smp_valid_read_clk      : std_logic;
  signal smp_cnt_read_clk        : smp_clk_cnt_s;
  signal smp_clk_active_read_clk : std_logic;
  signal read_first              : std_logic := '1';
  signal ref_clk_tc_read_clk     : std_logic;

begin
  -- Assign initial values depending on measurement accuracy selected
  ref_clk_cnt_max <= REF_CLK_CNT_MAX_MHZ when (freq_meas_type = FREQ_IN_MHZ) else
                     REF_CLK_CNT_MAX_KHZ when (freq_meas_type = FREQ_IN_KHZ) else
                     REF_CLK_CNT_MAX_HZ;

  -- ref_clk counter
  -- Produces ref_clk_tc every ref_clk_max counts
  ref_clk_cnt_p : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      if (ref_clk_cnt = REF_CLK_CNT_2) then
        ref_clk_tc  <= '1';
      else
        ref_clk_tc  <= '0';
      end if;
      if (ref_clk_tc = '1') then
        ref_clk_cnt <= ref_clk_cnt_max;
      else
        ref_clk_cnt <= ref_clk_cnt - 1;
      end if;
    end if;
  end process;

  -- Generate smp_clk_edge to indicate if we have active samp clock
  -- during measurement period. Use smp_valid_ref_clk to detect edges
  -- on smp_clk. Use ref_clk_tc to clear.
  ref_clk_det_p : process(ref_clk)
  begin
    if rising_edge(ref_clk) then
      if (ref_clk_tc = '1') then
        smp_clk_edge <= '0';
        smp_clk_active <= smp_clk_edge;
      end if;
      if (smp_valid_ref_clk = '1') then
        smp_clk_edge <= '1';
      end if;
    end if;
  end process;

  -- Transfer timer_reset pulse to smp_clk domain
  smp_toggle_pulse_i : toggle_pulse
  port map(
    clk_a   => ref_clk,
    in_a    => ref_clk_tc,
    clk_b   => smp_clk,
    out_b   => ref_clk_tc_smp_clk);

  -- smp_clk counter
  -- Output sampled on ref_clk_tc_smp_clk from ref_clk timer
  smp_cnt_p : process(smp_clk)
  begin
    if rising_edge(smp_clk) then
      if (ref_clk_tc_smp_clk = '1') then
        smp_cnt      <= SMP_CNT_1;
        smp_cnt_last <= smp_cnt;
        smp_valid    <= '1';
      else
        smp_cnt      <= smp_cnt + 1;
        smp_valid    <= '0';
      end if;
    end if;
  end process;

  -- Transfer smp_cnt_last data to read_clk domain
  toggle_pulse_data_i : toggle_pulse_data
  generic map(
    width   => 32)
  port map(
    clk_a   => smp_clk,
    send_a  => smp_valid,
    data_a  => smp_cnt_last,
    clk_b   => read_clk,
    data_b  => smp_cnt_read_clk,
    valid_b => smp_valid_read_clk);

  -- Transfer smp_valid pulse to ref_clk domain
  ref1_toggle_pulse_i : toggle_pulse
  port map(
    clk_a   => smp_clk,
    in_a    => smp_valid,
    clk_b   => ref_clk,
    out_b   => smp_valid_ref_clk);


  -- Transfer smp_clk_active pulse from ref_clk to read_clk domain
  -- smp_clk_active is only updated every ref_clk_tc
  active_sync2_i : syncn
  port map(
    clk   => read_clk,
    d     => smp_clk_active,
    q     => smp_clk_active_read_clk);


  -- Transfer ref_clk_tc pulse to read_clk domain
  ref2_toggle_pulse_i : toggle_pulse
  port map(
    clk_a   => ref_clk,
    in_a    => ref_clk_tc,
    clk_b   => read_clk,
    out_b   => ref_clk_tc_read_clk);

  -- Generate outputs in read_clk domain.
  -- Clear read_first after first measurement period is complete
  -- as we have flushed out the initisl invalid measurement.
  output_p : process(read_clk)
  begin
    if rising_edge(read_clk) then
      -- If smp_clk is running then use sampled frequency and smp_valid update
      if (smp_clk_active_read_clk = '1') then
        if (smp_valid_read_clk = '1') then
          read_first <= '0';
        end if;
        read_freq   <= smp_cnt_read_clk;
        read_update <= smp_valid_read_clk and not(read_first);
      -- If smp_clk is not running then use zero frequency and ref_clk_tc update
      else
        if (ref_clk_tc_read_clk = '1') then
          read_first <= '0';
        end if;
        read_freq   <= (others => '0');
        read_update <= ref_clk_tc_read_clk and not(read_first);
      end if;
    end if;
  end process;

  read_running <= smp_clk_active_read_clk;

end architecture;