--------------------------------------------------------------------------------
--
-- Package Name     : mptl
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          :
--
-- Modified By      :
-- Date             :
-- Change Notes     :
--
-- Description      : Constant definitions for MPTL Packet Interface
--
--
-- Dependencies     : adb3_ocp
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

package mptl is

  subtype mptl_packet_type is std_logic_vector(3 downto 0);
  constant MPTL_NO_PACKET         : mptl_packet_type := X"0";
  constant MPTL_SINGLE_BEAT_WRITE : mptl_packet_type := X"1";
  constant MPTL_READ_REQUEST      : mptl_packet_type := X"2";
  constant MPTL_BURST_WRITE_CMD   : mptl_packet_type := X"3";
  constant MPTL_BURST_WRITE_DATA  : mptl_packet_type := X"4";
  constant MPTL_READ_DATA         : mptl_packet_type := X"5";
  constant MPTL_STATUS            : mptl_packet_type := X"8";
  constant MPTL_COMMAND           : mptl_packet_type := X"9";

  type mptl64_data_array is array (natural range <>) of std_logic_vector(63 downto 0);
  type mptl128_data_array is array (natural range <>) of std_logic_vector(127 downto 0);

  constant MPTL_CONTROLLER_PACKET_SIZE : integer := 4;
  constant MPTL_SINGLE_BEAT_WRITE_PACKET_SIZE : integer := 4;
  constant MPTL_READ_REQUEST_PACKET_SIZE : integer := 2;
  constant MPTL_BURST_WRITE_CMD_PACKET_SIZE : integer := 2;
  constant MPTL_MAX_BURST_PAYLOAD_SIZE : integer := 16;
  constant MPTL_MAX_BURST_PAYLOAD_PACKET_SIZE : integer := MPTL_MAX_BURST_PAYLOAD_SIZE+1;
  constant MPTL128_MAX_BURST_PAYLOAD_SIZE : integer := 16;
  constant MPTL128_MAX_BURST_PAYLOAD_PACKET_SIZE : integer := MPTL128_MAX_BURST_PAYLOAD_SIZE+1;
  constant MPTL128_CONTROLLER_PACKET_SIZE : integer := 2;
  constant MPTL128_SINGLE_BEAT_WRITE_PACKET_SIZE : integer := 2;
  constant MPTL128_READ_REQUEST_PACKET_SIZE : integer := 1;
  constant MPTL128_BURST_WRITE_CMD_PACKET_SIZE : integer := 1;


  component adb3_ocp_to_mptl64
    generic (
      channel : integer := 0);
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : in  adb3_ocp_m2sT;
      ocp_s2m            : out adb3_ocp_s2mT;
      -- MPTL interface
      mptl64_data        : out std_logic_vector(63 downto 0);
      mptl64_valid       : out std_logic;
      mptl64_accept      : in  std_logic;
      online             : in  std_logic;
      -- Last packet consumed by
      current_packet_out : out std_logic_vector(7 downto 0);
      remote_last_packet : in  std_logic_vector(7 downto 0));
  end component;

  component adb3_ocp_resp_to_mptl64
    generic (
      channel : integer := 0);
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : out adb3_ocp_m2sT;
      ocp_s2m            : in  adb3_ocp_s2mT;
      -- MPTL interface
      mptl64_data        : out std_logic_vector(63 downto 0);
      mptl64_valid       : out std_logic;
      mptl64_accept      : in  std_logic;
      online             : in  std_logic;
      -- Last packet consumed by
      remote_last_packet : in  std_logic_vector(7 downto 0);
      current_packet_out : out  std_logic_vector(7 downto 0);
      -- Burst Length Info
      local_ocp_read     : in  std_logic;
      local_burst_length : in  std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
      local_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
      local_ocp_nfull    : out std_logic;  -- stop further OCP reads
      -- Error
      tag_mismatch       : out std_logic);
  end component;

  component mptl64_to_adb3_ocp
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : out adb3_ocp_m2sT;
      ocp_s2m            : in  adb3_ocp_s2mT;
      -- MPTL interface
      mptl64_data        : in  std_logic_vector(63 downto 0);
      mptl64_valid       : in  std_logic;
      online             : in  std_logic;
      -- Last read from FIFO
      remote_last_packet : out std_logic_vector(7 downto 0);
      -- Connection to OCP RESP for reads
      local_ocp_read     : out std_logic;
      local_burst_length : out std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
      local_tag          : out std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
      local_ocp_nfull    : in  std_logic);
  end component;

  component mptl64_to_adb3_ocp_resp
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : in  adb3_ocp_m2sT;
      ocp_s2m            : out adb3_ocp_s2mT;
      -- MPTL interface
      mptl64_data        : in  std_logic_vector(63 downto 0);
      mptl64_valid       : in  std_logic;
      online             : in  std_logic;
      -- Last read from FIFO
      remote_last_packet : out std_logic_vector(7 downto 0));
  end component;


  component mptl64_packet_mux
    generic (
      number_of_ports : natural := 5);
    port (
      -- OCP Interface
      clk               : in  std_logic;
      -- MPTL Multi-port input interface
      mptl64_data_in    : in  mptl64_data_array(0 to number_of_ports-1);
      mptl64_valid_in   : in  std_logic_vector(number_of_ports-1 downto 0);
      mptl64_accept_in  : out std_logic_vector(number_of_ports-1 downto 0);
      -- MPTL Single-port output
      mptl64_data_out   : out std_logic_vector(63 downto 0);
      mptl64_valid_out  : out std_logic;
      mptl64_fifo_nfull : in  std_logic);
  end component;


  component mptl64_packet_demux
    generic (
      number_of_ports : natural := 5);
    port (
      -- OCP Interface
      clk              : in  std_logic;
      -- MPTL Multi-port output interface
      mptl64_data_out  : out mptl64_data_array(0 to number_of_ports-1);
      mptl64_valid_out : out std_logic_vector(number_of_ports-1 downto 0);
      -- MPTL Single-port input
      mptl64_data_in   : in  std_logic_vector(63 downto 0);
      mptl64_valid_in  : in  std_logic;
      -- Error detection
      error_detected   : out std_logic);
  end component;


  component mptl64_packet_buffer is
    port (
      -- OCP Interface
      ocp_clk         : in  std_logic;
      -- MPTL Multi-port output interface
      mptl64_data_in  : in  std_logic_vector(63 downto 0);
      mptl64_valid_in : in  std_logic;
      mptl64_nfull    : out std_logic;
      -- GTP Outputs
      gtp_clk         : in  std_logic;
      gtp_data        : out std_logic_vector(63 downto 0);
      gtp_charisk     : out std_logic_vector(7 downto 0);
      gtp_online      : in  std_logic);
  end component;

  component mptl64_gtp_buffer is
    port (
      -- OCP Interface
      ocp_clk      : in  std_logic;
      -- MPTL Multi-port output interface
      mptl64_data  : out std_logic_vector(63 downto 0);
      mptl64_valid : out std_logic;
      -- GTP Outputs
      gtp_clk      : in  std_logic;
      gtp_data     : in  std_logic_vector(63 downto 0);
      gtp_charisk  : in  std_logic_vector(7 downto 0);
      gtp_online   : out std_logic;
      tx_offline   : in  std_logic);
  end component;


  component mptl64_controller
    generic (
      channel       : natural := 1;
      update_period : natural := 128);
    port (
      -- MPTL_Interfaces
      clk                 : in  std_logic;
      mptl64_data_out     : out std_logic_vector(63 downto 0)  := (others => '0');
      mptl64_valid_out    : out std_logic                      := '0';
      mptl64_accept_out   : in  std_logic;
      mptl64_data_in      : in  std_logic_vector(63 downto 0);
      mptl64_valid_in     : in  std_logic;
      -- Last Packet Levels
      local_last_packets  : in  std_logic_vector(127 downto 0);
      remote_last_packets : out std_logic_vector(127 downto 0) := (others => '0');
      -- GPIO (for interrupts?)
      gpio_in             : in  std_logic_vector(15 downto 0);
      gpio_out            : out std_logic_vector(15 downto 0)  := (others => '0');
      gpio_status         : in  std_Logic_vector(31 downto 0);
      remote_status       : out std_Logic_vector(31 downto 0);
      packet_sent         : out std_logic                      := '0';
      packet_received     : out std_logic                      := '0';
      online              : in  std_logic);
  end component;

  component mptl64_gtp_interface
    port (
      clkin           : in  std_logic;
      rst             : in  std_logic;
      ref_clk         : out std_logic;
      gtp_clk         : in  std_logic;
      gtp_clk2x       : in  std_logic;
      -- Connection to MPTL64HSS Buffers
      gtp_data_in     : in  std_logic_vector(63 downto 0);
      gtp_charisk_in  : in  std_logic_vector(7 downto 0);
      gtp_data_out    : out std_logic_vector(63 downto 0);
      gtp_charisk_out : out std_logic_vector(7 downto 0);
      -- Connection to Serial Ports
      tx_p            : out std_logic_vector(3 downto 0);
      tx_n            : out std_logic_vector(3 downto 0);
      rx_p            : in  std_logic_vector(3 downto 0);
      rx_n            : in  std_logic_vector(3 downto 0);
      -- Control
      loopback        : in  std_logic_vector(2 downto 0);
      rx_reset        : in  std_logic_vector(3 downto 0);
      en_comma_align  : in  std_logic;
      powerdown       : in  std_logic;
      rx_status       : out std_logic_vector(11 downto 0);
      rx_loss_of_sync : out std_logic_vector(7 downto 0);
      pll_locked      : out std_Logic;
      debug           : out std_Logic_vector(15 downto 0));
  end component;

  component mptl64_gtx_interface
    port (
      clkin           : in  std_logic;
      rst             : in  std_logic;
      ref_clk         : out std_logic;
      gtp_clk         : in  std_logic;
      gtp_clk2x       : in  std_logic;
      gtp_data_in     : in  std_logic_vector(63 downto 0);
      gtp_charisk_in  : in  std_Logic_vector(7 downto 0);
      gtp_data_out    : out std_logic_vector(63 downto 0);
      gtp_charisk_out : out std_Logic_vector(7 downto 0);
      tx_p            : out std_logic_vector(3 downto 0);
      tx_n            : out std_logic_vector(3 downto 0);
      rx_p            : in  std_logic_vector(3 downto 0);
      rx_n            : in  std_logic_vector(3 downto 0);
      loopback        : in  std_logic_vector(2 downto 0);
      rx_reset        : in  std_Logic_vector(3 downto 0);
      en_comma_align  : in  std_logic;
      powerdown       : in  std_Logic;
      rx_status       : out std_logic_vector(11 downto 0);
      rx_loss_of_sync : out std_logic_vector(7 downto 0);
      pll_locked      : out std_Logic);
  end component;


  component mptl64_gtx6_interface
    port (
      clkin           : in  std_logic;
      rst             : in  std_logic;
      ref_clk         : out std_logic;
      gtp_clk         : in  std_logic;
      gtp_clk2x       : in  std_logic;
      gtp_data_in     : in  std_logic_vector(63 downto 0);
      gtp_charisk_in  : in  std_logic_vector(7 downto 0);
      gtp_data_out    : out std_logic_vector(63 downto 0);
      gtp_charisk_out : out std_logic_vector(7 downto 0);
      tx_p            : out std_logic_vector(3 downto 0);
      tx_n            : out std_logic_vector(3 downto 0);
      rx_p            : in  std_logic_vector(3 downto 0);
      rx_n            : in  std_logic_vector(3 downto 0);
      loopback        : in  std_logic_vector(2 downto 0);
      rx_reset        : in  std_logic_vector(3 downto 0);
      en_comma_align  : in  std_logic;
      powerdown       : in  std_logic;
      rx_status       : out std_logic_vector(11 downto 0);
      rx_loss_of_sync : out std_logic_vector(7 downto 0);
      pll_locked      : out std_Logic;
      debug           : out std_Logic_vector(15 downto 0));
  end component;

  component mptl64_interface_bridge
    port (
      -- OCP Interface
      ocp_clk           : in  std_logic;
      direct_slave_m2s  : in  adb3_ocp_m2sT;
      direct_slave_s2m  : out adb3_ocp_s2mT;
      dma_channel0_m2s  : in  adb3_ocp_m2sT;
      dma_channel0_s2m  : out adb3_ocp_s2mT;
      dma_channel1_m2s  : in  adb3_ocp_m2sT;
      dma_channel1_s2m  : out adb3_ocp_s2mT;
      direct_master_m2s : out adb3_ocp_m2sT;
      direct_master_s2m : in  adb3_ocp_s2mT;
      -- GTP Interface
      gtp_clk_p         : in  std_logic;
      gtp_clk_n         : in  std_logic;
      tx_p              : out std_logic_vector(3 downto 0);
      tx_n              : out std_logic_vector(3 downto 0);
      rx_p              : in  std_logic_vector(3 downto 0);
      rx_n              : in  std_logic_vector(3 downto 0);
      -- Bridge to Target Sideband Signals
      target_configured : in  std_logic;
      target_gtp_online : in  std_logic;
      bridge_gtp_online : out std_logic;
      -- MPTL64 Sideband Interface
      gpio_in           : in  std_logic_vector(15 downto 0);
      gpio_out          : out std_logic_vector(15 downto 0);
      gtp_status        : out std_logic_vector(31 downto 0));
  end component;

  component mptl64_interface_target
    port (
      -- OCP Interface
      ocp_clk           : in  std_logic;
      direct_slave_m2s  : out adb3_ocp_m2sT;
      direct_slave_s2m  : in  adb3_ocp_s2mT;
      dma_channel0_m2s  : out adb3_ocp_m2sT;
      dma_channel0_s2m  : in  adb3_ocp_s2mT;
      dma_channel1_m2s  : out adb3_ocp_m2sT;
      dma_channel1_s2m  : in  adb3_ocp_s2mT;
      direct_master_m2s : in  adb3_ocp_m2sT;
      direct_master_s2m : out adb3_ocp_s2mT;
      -- GTP Interface
      gtp_clk_p         : in  std_logic;
      gtp_clk_n         : in  std_logic;
      tx_p              : out std_logic_vector(3 downto 0);
      tx_n              : out std_logic_vector(3 downto 0);
      rx_p              : in  std_logic_vector(3 downto 0);
      rx_n              : in  std_logic_vector(3 downto 0);
      -- Bridge to Target Sideband Signals
      target_configured : out std_logic;
      target_gtp_online : out std_logic;
      bridge_gtp_online : in  std_logic;
      -- MPTL64 Sideband Interface
      gpio_in           : in  std_logic_vector(15 downto 0);
      gpio_out          : out std_logic_vector(15 downto 0));
  end component;


  procedure mptl64_read_request (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant burst_length : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer);


  procedure mptl64_single_write (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant data         : in  integer;
    constant byte_enables : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer);

  procedure mptl64_burst_write_cmd (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant burst_length : in  integer;
    constant byte_enables : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer);

  procedure mptl64_burst_write_data (
    signal clk             : in  std_logic;
    signal mptl64_data     : out std_logic_vector(63 downto 0);
    signal mptl64_valid    : out std_logic;
    signal mptl64_accept   : in  std_logic;
    constant first_data    : in  integer;
    constant packet_length : in  integer;
    constant channel       : in  integer;
    constant packet_id     : in  integer);

  procedure mptl64_read_data (
    signal clk             : in  std_logic;
    signal mptl64_data     : out std_logic_vector(63 downto 0);
    signal mptl64_valid    : out std_logic;
    signal mptl64_accept   : in  std_logic;
    constant first_data    : in  integer;
    constant packet_length : in  integer;
    constant tag           : in  integer;
    constant channel       : in  integer;
    constant packet_id     : in  integer);


  component adb3_ocp_to_mptl128
    generic (
      channel : integer := 0);
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : in  adb3_ocp_m2sT;
      ocp_s2m            : out adb3_ocp_s2mT;
      -- MPTL interface
      mptl128_data        : out std_logic_vector(127 downto 0);
      mptl128_valid       : out std_logic;
      mptl128_accept      : in  std_logic;
      online             : in  std_logic;
      -- Last packet consumed by
      remote_last_packet : in  std_logic_vector(7 downto 0));
  end component;

  component adb3_ocp_resp_to_mptl128
    generic (
      channel : integer := 0);
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : out adb3_ocp_m2sT;
      ocp_s2m            : in  adb3_ocp_s2mT;
      -- MPTL interface
      mptl128_data        : out std_logic_vector(127 downto 0);
      mptl128_valid       : out std_logic;
      mptl128_accept      : in  std_logic;
      online             : in  std_logic;
      -- Last packet consumed by
      remote_last_packet : in  std_logic_vector(7 downto 0);
      -- Burst Length Info
      local_ocp_read     : in  std_logic;
      local_burst_length : in  std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
      local_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
      local_ocp_nfull    : out std_logic;  -- stop further OCP reads
      -- Error
      tag_mismatch       : out std_logic);
  end component;

  component mptl128_to_adb3_ocp
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : out adb3_ocp_m2sT;
      ocp_s2m            : in  adb3_ocp_s2mT;
      -- MPTL interface
      mptl128_data        : in  std_logic_vector(127 downto 0);
      mptl128_valid       : in  std_logic;
      online             : in  std_logic;
      -- Last read from FIFO
      remote_last_packet : out std_logic_vector(7 downto 0);
      -- Connection to OCP RESP for reads
      local_ocp_read     : out std_logic;
      local_burst_length : out std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
      local_tag          : out std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
      local_ocp_nfull    : in  std_logic);
  end component;

  component mptl128_to_adb3_ocp_resp
    port (
      -- OCP Interface
      ocp_clk            : in  std_logic;
      ocp_m2s            : in  adb3_ocp_m2sT;
      ocp_s2m            : out adb3_ocp_s2mT;
      -- MPTL interface
      mptl128_data        : in  std_logic_vector(127 downto 0);
      mptl128_valid       : in  std_logic;
      online             : in  std_logic;
--       debug : out std_logic_vector(7 downto 0);
      -- Last read from FIFO
      remote_last_packet : out std_logic_vector(7 downto 0));

  end component;


  component mptl128_packet_mux
    generic (
      number_of_ports : natural := 5);
    port (
      -- OCP Interface
      clk               : in  std_logic;
      online            : in  std_logic;
      -- MPTL Multi-port input interface
      mptl128_data_in    : in  mptl128_data_array(0 to number_of_ports-1);
      mptl128_valid_in   : in  std_logic_vector(number_of_ports-1 downto 0);
      mptl128_accept_in  : out std_logic_vector(number_of_ports-1 downto 0);
      -- MPTL Single-port output
      mptl128_data_out   : out std_logic_vector(127 downto 0);
      mptl128_valid_out  : out std_logic;
      mptl128_fifo_nfull : in  std_logic);
  end component;


  component mptl128_packet_demux
    generic (
      number_of_ports : natural := 5);
    port (
      -- OCP Interface
      clk              : in  std_logic;
      online            : in  std_logic;
      -- MPTL Multi-port output interface
      mptl128_data_out  : out mptl128_data_array(0 to number_of_ports-1);
      mptl128_valid_out : out std_logic_vector(number_of_ports-1 downto 0);
      -- MPTL Single-port input
      mptl128_data_in   : in  std_logic_vector(127 downto 0);
      mptl128_valid_in  : in  std_logic;
      -- Error detection
      error_detected   : out std_logic);
  end component;


  component mptl128_packet_buffer is
    port (
      -- OCP Interface
      ocp_clk         : in  std_logic;
      -- MPTL Multi-port output interface
      mptl128_data_in  : in  std_logic_vector(127 downto 0);
      mptl128_valid_in : in  std_logic;
      mptl128_nfull    : out std_logic;
      mptl_retry       : in  std_logic;
      -- GTP Outputs
      gtp_clk         : in  std_logic;
      gtp_data        : out std_logic_vector(127 downto 0);
      gtp_charisk     : out std_logic_vector(15 downto 0);
      gtp_online      : in  std_logic);
  end component;

  component mptl128_gtp_buffer is
    generic (
      four_gtps : boolean := true);
    port (
      -- OCP Interface
      ocp_clk      : in  std_logic;
      -- MPTL Multi-port output interface
      mptl128_data  : out std_logic_vector(127 downto 0);
      mptl128_valid : out std_logic;
      -- GTP Outputs
      gtp_clk      : in  std_logic;
      gtp_data     : in  std_logic_vector(127 downto 0);
      gtp_charisk  : in  std_logic_vector(15 downto 0);
      gtp_online   : out std_logic;
      tx_offline   : in  std_logic;
      error8b10b    : in  std_logic;
      mptl_retry    : out std_logic);
  end component;


  component mptl64par_packet_buffer
    port (
      ocp_clk         : in  std_logic;
      mptl64_data_in  : in  std_logic_vector(63 downto 0);
      mptl64_valid_in : in  std_logic;
      mptl64_nfull    : out std_logic;
      gtp_clk         : in  std_logic;
      tx_d            : out std_logic_vector(31 downto 0);
      tx_v            : out std_logic;
      tx_ck           : out std_logic_vector(1 downto 0));
  end component;

  component mptl64par_buffer
    port (
      ocp_clk      : in  std_logic;
      mptl64_data  : out std_logic_vector(63 downto 0);
      mptl64_valid : out std_logic;
      debug        : out std_logic_vector(65 downto 0);
      rx_d         : in  std_logic_vector(31 downto 0);
      rx_v         : in  std_logic;
      rx_ck        : in  std_logic_vector(1 downto 0));
  end component;

  component mptl128_controller
    generic (
      channel       : natural := 1;
      update_period : natural := 128);
    port (
      -- MPTL_Interfaces
      clk                 : in  std_logic;
      mptl128_data_out     : out std_logic_vector(127 downto 0)  := (others => '0');
      mptl128_valid_out    : out std_logic                      := '0';
      mptl128_accept_out   : in  std_logic;
      mptl128_data_in      : in  std_logic_vector(127 downto 0);
      mptl128_valid_in     : in  std_logic;
      -- Last Packet Levels
      local_last_packets  : in  std_logic_vector(127 downto 0);
      remote_last_packets : out std_logic_vector(127 downto 0) := (others => '0');
      -- GPIO (for interrupts?)
      gpio_in             : in  std_logic_vector(31 downto 0);
      gpio_out            : out std_logic_vector(31 downto 0)  := (others => '0');
      gpio_status         : in  std_Logic_vector(31 downto 0);
      remote_status       : out std_Logic_vector(31 downto 0);
      packet_sent         : out std_logic                      := '0';
      packet_received     : out std_logic                      := '0';
      online              : in  std_logic);
  end component;

  component mptl128_gtp_interface
    port (
      clkin           : in  std_logic;
      rst             : in  std_logic;
      ref_clk         : out std_logic;
      gtp_clk         : in  std_logic;
      gtp_clk2x       : in  std_logic;
      -- Connection to MPTL64HSS Buffers
      gtp_data_in     : in  std_logic_vector(127 downto 0);
      gtp_charisk_in  : in  std_logic_vector(15 downto 0);
      gtp_data_out    : out std_logic_vector(127 downto 0);
      gtp_charisk_out : out std_logic_vector(15 downto 0);
      -- Connection to Serial Ports
      tx_p            : out std_logic_vector(7 downto 0);
      tx_n            : out std_logic_vector(7 downto 0);
      rx_p            : in  std_logic_vector(7 downto 0);
      rx_n            : in  std_logic_vector(7 downto 0);
      -- Control
      loopback        : in  std_logic_vector(2 downto 0);
      rx_reset        : in  std_logic_vector(7 downto 0);
      en_comma_align  : in  std_logic;
      powerdown       : in  std_logic;
      rx_status       : out std_logic_vector(23 downto 0);
      rx_loss_of_sync : out std_logic_vector(15 downto 0));
  end component;

  component mptl128_gtx6_interface is
    port (
      clkin           : in  std_logic;
      rst             : in  std_logic;
      ref_clk         : out std_logic;
      gtp_clk         : in  std_logic;
      gtp_clk2x       : in  std_logic;
      gtp_data_in     : in  std_logic_vector(127 downto 0);
      gtp_charisk_in  : in  std_logic_vector(15 downto 0);
      gtp_data_out    : out std_logic_vector(127 downto 0);
      gtp_charisk_out : out std_logic_vector(15 downto 0);
      tx_p            : out std_logic_vector(3 downto 0);
      tx_n            : out std_logic_vector(3 downto 0);
      rx_p            : in  std_logic_vector(3 downto 0);
      rx_n            : in  std_logic_vector(3 downto 0);
      loopback        : in  std_logic_vector(2 downto 0);
      config          : in  std_Logic_vector(15 downto 0);
      rx_reset        : in  std_logic_vector(7 downto 0);
      en_comma_align  : in  std_logic;
      powerdown       : in  std_logic;
      pll_locked      : out std_logic;
      rx_status       : out std_logic_vector(23 downto 0);
      rx_loss_of_sync : out std_logic_vector(15 downto 0));
  end component mptl128_gtx6_interface;




end mptl;

package body mptl is

  procedure mptl64_read_request (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant burst_length : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer)
  is
    variable address : std_logic_vector(63 downto 0);
  begin
    wait until clk = '1';
    wait until clk = '0';
    address := CONV_STD_LOGIC_VECTOR(addr, 64);

    mptl64_data(3 downto 0)   <= MPTL_READ_REQUEST;
    mptl64_data(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);
    mptl64_data(15 downto 8)  <= CONV_STD_LOGIC_VECTOR(2, 8);
    mptl64_data(23 downto 16) <= CONV_STD_LOGIC_VECTOR(packet_id, 8);
    mptl64_data(35 downto 24) <= CONV_STD_LOGIC_VECTOR(burst_length, 12);
    mptl64_data(39 downto 36) <= (others => '0');
    mptl64_data(47 downto 40) <= CONV_STD_LOGIC_VECTOR(tag, 8);
    mptl64_data(63 downto 48) <= address(63 downto 48);
    mptl64_valid              <= '1';
    wait until clk = '1';
    while mptl64_accept = '0' loop
      wait until clk = '0';
      wait until clk = '1';
    end loop;
    wait until clk = '0';
    mptl64_data(47 downto 0)  <= address(47 downto 0);
    mptl64_data(63 downto 48) <= (others => '0');
    wait until clk = '1';
    wait until clk = '0';
    mptl64_valid              <= '0';
  end;


  procedure mptl64_single_write (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant data         : in  integer;
    constant byte_enables : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer)
  is
    variable address : std_logic_vector(63 downto 0);
  begin
    wait until clk = '1';
    wait until clk = '0';
    address := CONV_STD_LOGIC_VECTOR(addr, 64);

    mptl64_data(3 downto 0)   <= MPTL_SINGLE_BEAT_WRITE;
    mptl64_data(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);
    mptl64_data(15 downto 8)  <= CONV_STD_LOGIC_VECTOR(4, 8);
    mptl64_data(23 downto 16) <= CONV_STD_LOGIC_VECTOR(packet_id, 8);
    mptl64_data(35 downto 24) <= CONV_STD_LOGIC_VECTOR(1, 12);
    mptl64_data(39 downto 36) <= (others => '0');
    mptl64_data(47 downto 40) <= CONV_STD_LOGIC_VECTOR(tag, 8);
    mptl64_data(63 downto 48) <= address(63 downto 48);
    mptl64_valid              <= '1';
    wait until clk = '1';
    while mptl64_accept = '0' loop
      wait until clk = '0';
      wait until clk = '1';
    end loop;
    wait until clk = '0';
    mptl64_data(47 downto 0)  <= address(47 downto 0);
    mptl64_data(63 downto 48) <= CONV_STD_LOGIC_VECTOR(byte_enables, 16);
    wait until clk = '1';
    wait until clk = '0';
    mptl64_data               <= CONV_STD_LOGIC_VECTOR(data, 64);
    wait until clk = '1';
    wait until clk = '0';
    mptl64_data               <= CONV_STD_LOGIC_VECTOR(data+1, 64);
    wait until clk = '1';
    wait until clk = '0';
    mptl64_valid              <= '0';
  end;

  procedure mptl64_burst_write_cmd (
    signal clk            : in  std_logic;
    signal mptl64_data    : out std_logic_vector(63 downto 0);
    signal mptl64_valid   : out std_logic;
    signal mptl64_accept  : in  std_logic;
    constant addr         : in  integer;
    constant burst_length : in  integer;
    constant byte_enables : in  integer;
    constant tag          : in  integer;
    constant channel      : in  integer;
    constant packet_id    : in  integer)
  is
    variable address : std_logic_vector(63 downto 0);
  begin
    wait until clk = '1';
    wait until clk = '0';
    address := CONV_STD_LOGIC_VECTOR(addr, 64);

    mptl64_data(3 downto 0)   <= MPTL_BURST_WRITE_CMD;
    mptl64_data(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);
    mptl64_data(15 downto 8)  <= CONV_STD_LOGIC_VECTOR(2, 8);
    mptl64_data(23 downto 16) <= CONV_STD_LOGIC_VECTOR(packet_id, 8);
    mptl64_data(35 downto 24) <= CONV_STD_LOGIC_VECTOR(burst_length, 12);
    mptl64_data(39 downto 36) <= (others => '0');
    mptl64_data(47 downto 40) <= CONV_STD_LOGIC_VECTOR(tag, 8);
    mptl64_data(63 downto 48) <= address(63 downto 48);
    mptl64_valid              <= '1';
    wait until clk = '1';
    while mptl64_accept = '0' loop
      wait until clk = '0';
      wait until clk = '1';
    end loop;
    wait until clk = '0';
    mptl64_data(47 downto 0)  <= address(47 downto 0);
    mptl64_data(63 downto 48) <= CONV_STD_LOGIC_VECTOR(byte_enables, 16);
    wait until clk = '1';
    wait until clk = '0';
    mptl64_valid              <= '0';
  end;

  procedure mptl64_burst_write_data (
    signal clk             : in  std_logic;
    signal mptl64_data     : out std_logic_vector(63 downto 0);
    signal mptl64_valid    : out std_logic;
    signal mptl64_accept   : in  std_logic;
    constant first_data    : in  integer;
    constant packet_length : in  integer;
    constant channel       : in  integer;
    constant packet_id     : in  integer)
  is
    variable packet_count : integer;
    variable data         : integer;
  begin
    wait until clk = '1';
    wait until clk = '0';
    packet_count := 1;
    data         := first_data;

    mptl64_data(3 downto 0)   <= MPTL_BURST_WRITE_DATA;
    mptl64_data(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);
    mptl64_data(15 downto 8)  <= CONV_STD_LOGIC_VECTOR(packet_length, 8);
    mptl64_data(23 downto 16) <= CONV_STD_LOGIC_VECTOR(packet_id, 8);
    mptl64_data(63 downto 24) <= (others => '0');
    mptl64_valid              <= '1';
    wait until clk = '1';
    while mptl64_accept = '0' loop
      wait until clk = '0';
      wait until clk = '1';
    end loop;
    wait until clk = '0';
    while (packet_count < packet_length) loop
      mptl64_data  <= CONV_STD_LOGIC_VECTOR(data, 64);
      wait until clk = '1';
      wait until clk = '0';
      data         := data+1;
      packet_count := packet_count+1;
    end loop;
    mptl64_valid <= '0';
  end;


  procedure mptl64_read_data (
    signal clk             : in  std_logic;
    signal mptl64_data     : out std_logic_vector(63 downto 0);
    signal mptl64_valid    : out std_logic;
    signal mptl64_accept   : in  std_logic;
    constant first_data    : in  integer;
    constant packet_length : in  integer;
    constant tag           : in  integer;
    constant channel       : in  integer;
    constant packet_id     : in  integer)
  is
    variable packet_count : integer;
    variable data         : integer;
  begin
    wait until clk = '1';
    wait until clk = '0';
    packet_count := 1;
    data         := first_data;

    mptl64_data(3 downto 0)   <= MPTL_READ_DATA;
    mptl64_data(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);
    mptl64_data(15 downto 8)  <= CONV_STD_LOGIC_VECTOR(packet_length, 8);
    mptl64_data(23 downto 16) <= CONV_STD_LOGIC_VECTOR(packet_id, 8);
    mptl64_data(39 downto 24) <= (others => '0');
    mptl64_data(47 downto 40) <= CONV_STD_LOGIC_VECTOR(tag, 8);
    mptl64_data(63 downto 48) <= (others => '0');
    mptl64_valid              <= '1';
    wait until clk = '1';
    while mptl64_accept = '0' loop
      wait until clk = '0';
      wait until clk = '1';
    end loop;
    wait until clk = '0';
    mptl64_valid              <= '0';
    wait until clk = '1';
    wait until clk = '0';
    wait until clk = '1';
      wait until clk = '0';
    mptl64_valid              <= '1';
    while (packet_count < packet_length) loop
      mptl64_data  <= CONV_STD_LOGIC_VECTOR(data, 64);
      wait until clk = '1';
      wait until clk = '0';
      mptl64_valid              <= '0';
      wait until clk = '1';
      wait until clk = '0';
      wait until clk = '1';
      wait until clk = '0';
      mptl64_valid              <= '1';
      data         := data+1;
      packet_count := packet_count+1;
    end loop;
    mptl64_valid <= '0';
  end;


end mptl;
