--==============================================================================
--
--     File Name: blk_clocks.vhd
--       Version: 0.1
--          Date: 12 Mar 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Clock generation block
--
--   Description: Connection and generation of clocks
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 12 Mar 2010: Initial design
--                0.2, KDC  08 Apr 2010: Changed PLL_BASE to MMCM_BASE
--                0.3, KDC  16 Sep 2010: Added pll_mem_rst/pll_mem_clk
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

library adcommon;
use adcommon.rst_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity blk_clocks is
port(
  -- Clock Inputs
  clks_non_mgt   : in    clks_non_mgt_t;
  clks_mgt       : in    clks_mgt_t;
  -- Reset Outputs
  pll_ref_rst    : out   std_logic;
  pll_reg_rst    : out   std_logic;
  pll_mem_rst    : out   std_logic;
  -- Clock Outputs
  pll_ref_clk    : out   std_logic;
  pll_reg_clk    : out   std_logic;
  pll_mem_clk    : out   std_logic;
  pll_mem_x2_clk : out   std_logic;
  clk_vec        : out   clk_vec_t;
  mptl_clk       : out   mptl_clk_t);
end;


architecture synth of blk_clocks is

  -- Reference clock signals
  signal ref_clk_ibuf   : std_logic;
  signal ref_clk_bufg   : std_logic;

  -- PLL signals
  signal pll_rst_count  : std_logic_vector(4 downto 0) := "11111";
  signal pll_rst        : std_logic := '1';
  signal fb_clk         : std_logic;
  signal fb_clk_bufg    : std_logic;
  signal pll_locked     : std_logic;
  signal pll_not_locked : std_logic;
  signal out_clk0       : std_logic;
  signal out_clk1       : std_logic;
  signal out_clk2       : std_logic;
  signal out_clk3       : std_logic;
  signal out_clk0_bufg  : std_logic;
  signal out_clk1_bufg  : std_logic;
  signal out_clk2_bufg  : std_logic;
  signal out_clk3_bufg  : std_logic;

  -- Non MGT Clock signals
  signal xrm_clkin_ibuf : std_logic;

  -- MGT clock signals
  signal mgt_clk_pin    : mgt_clk_pin_t;
  signal mgt_clk_ibuf   : mgt_clk_buf_t;
  signal clk_vec_bufg   : clk_vec_t;


begin

  ----------------------------------
  -- Design reference clock input --
  ----------------------------------

  -- Use 200MHz oscillator as design reference clock
  ref_clk_dif_i : IBUFDS
  port map(
    I  => clks_non_mgt.ref_clk.p,
    IB => clks_non_mgt.ref_clk.n,
    O  => ref_clk_ibuf);

  ref_clk_bufg_i : BUFG
  port map(
    I => ref_clk_ibuf,
    O => ref_clk_bufg);


  -- Generate PLL reset using buffered reference clock
  pll_rst_p : process(ref_clk_bufg)
  begin
    if rising_edge(ref_clk_bufg) then
      if pll_rst_count = "00000" then
        pll_rst       <= '0';
      else
        pll_rst       <= '1';
        pll_rst_count <= pll_rst_count-1;
      end if;
    end if;
  end process;


  -- Reference clock MMCM (CLKFBOUT range 600 MHz to 1200 MHz)
  -- CLKFBOUT = (CLKIN1/DIVCLK_DIVIDE) * CLKFBOUT_MULT_F
  -- CLKOUTn  = (CLKIN1/DIVCLK_DIVIDE) * CLKFBOUT_MULT_F/CLKOUTn_DIVIDE
  -- CLKFBOUT = (200 MHz/2) * 12.000       = 1200 MHz
  -- CLKOUT0  = (200 MHz/2) * 12.000/6.000 =  200 MHz (pll_ref_clk)
  -- CLKOUT1  = (200 MHz/2) * 12.000/15    =   80 MHz (pll_reg_clk)
  -- CLKOUT2  = (200 MHz/2) * 12.000/6     =  200 MHz (pll_mem_clk)
  -- CLKOUT3  = (200 MHz/2) * 12.000/3     =  400 MHz (pll_mem_x2_clk)
  mmcm_ref_clk_i : MMCM_BASE
  generic map(
    BANDWIDTH          => "OPTIMIZED", -- string := "OPTIMIZED"
    CLKIN1_PERIOD      => 5.000,       -- real := 0.0
    DIVCLK_DIVIDE      => 2,           -- integer := 1 (1 to 128)
    CLKFBOUT_MULT_F    => 12.000,      -- real := 1.0  (5.0 to 64.0)
    CLKOUT0_DIVIDE_F   => 6.000,       -- real := 1.0  (1.0 to 128.0)
    CLKOUT1_DIVIDE     => 15,          -- integer := 1
    CLKOUT2_DIVIDE     => 6,           -- integer := 1
    CLKOUT3_DIVIDE     => 3,           -- integer := 1
    CLKOUT4_DIVIDE     => 6,           -- integer := 1
    CLKOUT5_DIVIDE     => 6,           -- integer := 1
    CLKOUT6_DIVIDE     => 6,           -- integer := 1
    CLKFBOUT_PHASE     => 0.000,       -- real := 0.0
    CLKOUT0_PHASE      => 0.000,       -- real := 0.0
    CLKOUT1_PHASE      => 0.000,       -- real := 0.0
    CLKOUT2_PHASE      => 0.000,       -- real := 0.0
    CLKOUT3_PHASE      => 0.000,       -- real := 0.0
    CLKOUT4_PHASE      => 0.000,       -- real := 0.0
    CLKOUT5_PHASE      => 0.000,       -- real := 0.0
    CLKOUT6_PHASE      => 0.000,       -- real := 0.0
    CLKOUT0_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT1_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT2_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT3_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT4_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT5_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT6_DUTY_CYCLE => 0.500,       -- real := 0.5
    CLKOUT4_CASCADE    => FALSE,       -- boolean := FALSE
    CLOCK_HOLD         => FALSE,       -- boolean := FALSE
    REF_JITTER1        => 0.0,         -- real := 0.0
    STARTUP_WAIT       => FALSE)       -- boolean := FALSE
  port map(
    RST       => pll_rst,      -- in std_ulogic;
    PWRDWN    => '0',          -- in std_ulogic;
    CLKIN1    => ref_clk_bufg, -- in std_ulogic;
    CLKFBIN   => fb_clk_bufg,  -- in std_ulogic;
    CLKFBOUT  => fb_clk,       -- out std_ulogic;
    CLKFBOUTB => open,         -- out std_ulogic;
    CLKOUT0   => out_clk0,     -- out std_ulogic;
    CLKOUT0B  => open,         -- out std_ulogic;
    CLKOUT1   => out_clk1,     -- out std_ulogic;
    CLKOUT1B  => open,         -- out std_ulogic;
    CLKOUT2   => out_clk2,     -- out std_ulogic;
    CLKOUT2B  => open,         -- out std_ulogic;
    CLKOUT3   => out_clk3,     -- out std_ulogic;
    CLKOUT3B  => open,         -- out std_ulogic;
    CLKOUT4   => open,         -- out std_ulogic;
    CLKOUT5   => open,         -- out std_ulogic;
    CLKOUT6   => open,         -- out std_ulogic;
    LOCKED    => pll_locked);  -- out std_ulogic;

  pll_not_locked <= not(pll_locked);

  -- MMCM feedback clock
  fb_clk_bufg_i : BUFG
  port map(
    I => fb_clk,
    O => fb_clk_bufg);

  -- MMCM output clock0 => pll_ref_clk internal reference clock
  out_clk0_bufg_i : BUFG
  port map(
    I => out_clk0,
    O => out_clk0_bufg);

  pll_ref_clk <= out_clk0_bufg;

  -- MMCM output clock1 => pll_reg_clk register interface clock
  out_clk1_bufg_i : BUFG
  port map(
    I => out_clk1,
    O => out_clk1_bufg);

  pll_reg_clk <= out_clk1_bufg;

  -- MMCM output clock2 => pll_mem_clk memory interface clock
  out_clk2_bufg_i : BUFG
  port map(
    I => out_clk2,
    O => out_clk2_bufg);

  pll_mem_clk <= out_clk2_bufg;

  -- MMCM output clock3 => pll_mem_x2_clk memory interface clock
  out_clk3_bufg_i : BUFG
  port map(
    I => out_clk3,
    O => out_clk3_bufg);

  pll_mem_x2_clk <= out_clk3_bufg;

  -- Add more user clocks here
--  out_clk4_bufg_i : BUFG
--  port map(
--    I => out_clk4,
--    O => pll_usr4_clk);

--  pll_usr4_clk <= out_clk4_bufg;


  -- Generate async on, sync off, reset pll_ref_rst
  out_clk0_rst_i : rst_sync
  port map(
    async_rst  => pll_not_locked, -- in    std_logic
    clk        => out_clk0_bufg,  -- in    std_logic
    master_rst => pll_ref_rst);   -- out   std_logic);

  -- Generate async on, sync off, reset pll_reg_rst
  out_clk1_rst_i : rst_sync
  port map(
    async_rst  => pll_not_locked, -- in    std_logic
    clk        => out_clk1_bufg,  -- in    std_logic
    master_rst => pll_reg_rst);   -- out   std_logic);

  -- Generate async on, sync off, reset pll_mem_rst
  out_clk2_rst_i : rst_sync
  port map(
    async_rst  => pll_not_locked, -- in    std_logic
    clk        => out_clk2_bufg,  -- in    std_logic
    master_rst => pll_mem_rst);   -- out   std_logic);

  -- Add more user clock resets here
--  out_clk3_rst_i : rst_sync
--  port map(
--    async_rst  => pll_not_locked, -- in    std_logic
--    clk        => out_clk3_bufg,  -- in    std_logic
--    master_rst => pll_usr3_rst);  -- out   std_logic);


  ----------------
  -- MGT clocks --
  ----------------

  -- Connect clks_mgt to mgt_clk_pin vector
  mgt_clk_pin(MGT112_CLK0_NUM) <= clks_mgt.mgt112_clk0;
  mgt_clk_pin(MGT113_CLK0_NUM) <= clks_mgt.mgt113_clk0;
  mgt_clk_pin(MGT113_CLK1_NUM) <= clks_mgt.mgt113_clk1;
  mgt_clk_pin(MGT114_CLK0_NUM) <= clks_mgt.mgt114_clk0;
  mgt_clk_pin(MGT115_CLK0_NUM) <= clks_mgt.mgt115_clk0;
  mgt_clk_pin(MGT116_CLK0_NUM) <= clks_mgt.mgt116_clk0;
  mgt_clk_pin(MGT117_CLK0_NUM) <= clks_mgt.mgt117_clk0;


  mgt_clk_loop_g : for n in MGT110_CLK0_NUM to MGT118_CLK1_NUM generate
    ibufds_gtxe1_g : if ((BOARD_TYPE = ADM_XRC_6TL)and(CLK_EN_6TL(n) = '1')) or
                        ((BOARD_TYPE = ADM_XRC_6T1)and(CLK_EN_6T1(n) = '1')) generate
      -- Extract mgt clks and buffer
      mgt_clk_ibufds_gtxe1_i : IBUFDS_GTXE1
      port map(
        I     => mgt_clk_pin(n).p,
        IB    => mgt_clk_pin(n).n,
        CEB   => '0',
        O     => mgt_clk_ibuf(n),
        ODIV2 => open);

      -- Buffer MGT clocks and connect to clk_vec_bufg vector
      mgt_clk_bufg_i : BUFG
      port map(
        I => mgt_clk_ibuf(n),
        O => clk_vec_bufg(n));
    end generate;
  end generate;


  --------------------
  -- Non MGT clocks --
  --------------------

  clk_vec_bufg_lclk_s_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- Buffer lclk_s and connect to clk_vec_bufg vector
    lclk_s_bufg_i : BUFG
    port map(
      I => clks_non_mgt.lclk,
      O => clk_vec_bufg(LCLK_NUM));
  end generate;

  clk_vec_bufg_xrm_g : if (BOARD_TYPE = ADM_XRC_6TL) or
                          (BOARD_TYPE = ADM_XRC_6T1) generate
    -- Extract xrm_clkin and buffer
    xrm_clk_dif_i : IBUFDS
    port map(
      I  => clks_non_mgt.xrm_clkin.p,
      IB => clks_non_mgt.xrm_clkin.n,
      O  => xrm_clkin_ibuf);

    -- Buffer xrm_clkin and connect to clk_vec_bufg vector
    xrm_clk_bufg_i : BUFG
    port map(
      I => xrm_clkin_ibuf,
      O => clk_vec_bufg(XRM_CLKIN_NUM));
  end generate;

  clk_vec <= clk_vec_bufg;


  ----------------
  -- MPTL clock --
  ----------------

  clk_vec_bufg_6tl_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- MPTL Parallel interface: Connect 250MHz PCIe input to mptl_clk
    mptl_clk.p <= clks_mgt.mgt114_clk0.p;
    mptl_clk.n <= clks_mgt.mgt114_clk0.n;
  end generate;

  clk_vec_bufg_6t1_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    -- MPTL serial interface: Connect 250 MHz Osc input to mptl_clk
    mptl_clk.p <= clks_mgt.mgt115_clk0.p;
    mptl_clk.n <= clks_mgt.mgt115_clk0.n;
  end generate;

end;