--==============================================================================
--
--     File Name: blk_ds_clk_read.vhd
--       Version: 0.1
--          Date: 16 Apr 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Clock frequency measurement
--
--   Description: Clock frequency measurement using clock_speed blocks
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 16 Apr 2010: Initial design
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.uber_pkg.all;

library adcommon;
use adcommon.clock_speed_pkg.all;

entity blk_ds_clk_read is
generic (
  addr_width : natural range 1 to DS_ADDR_WIDTH);
port(
  -- OCP Interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- Clock inputs
  pll_ref_clk      : in    std_logic;
  pll_reg_clk      : in    std_logic;
  pll_mem_clk      : in    std_logic;
  clk_vec          : in    clk_vec_t);
end;


architecture synth of blk_ds_clk_read is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  signal pll_ref_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_ref_clk_update  : std_logic;
  signal pll_ref_clk_running : std_logic;

  signal pll_reg_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_reg_clk_update  : std_logic;
  signal pll_reg_clk_running : std_logic;

  signal pll_mem_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_mem_clk_update  : std_logic;
  signal pll_mem_clk_running : std_logic;

  signal clk_freq            : clk_vec_freq_t;
  signal clk_update          : clk_vec_t;
  signal clk_running         : clk_vec_t;

  -- Clock frequency measurement registers
  signal ctrl_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal stat_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal sel_reg         : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ctrl_clr_update : std_logic;
  signal ctrl_reg_wr     : std_logic_vector(WE_WIDTH-1 downto 0);
  signal ctrl_wr         : std_logic;
  signal sel_clk         : std_logic_vector(4 downto 0);
  signal freq_reg        : std_logic_vector(REG_WIDTH-1 downto 0);

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- Clock frequency measurement
  signal pll_ref_clk_update_sr : std_logic;
  signal pll_reg_clk_update_sr : std_logic;
  signal pll_mem_clk_update_sr : std_logic;
  signal clk_update_sr         : clk_vec_t;

begin

  ---------------------------------
  -- Clock frequency measurement --
  ---------------------------------

  -- Measure pll_ref_clk frequency
  -- Use pll_ref_clk as reference clock frequency
  -- Use ocp_clk as read clock frequency
  pll_ref_clk_clock_speed_i : clock_speed
  generic map(
    ref_clk_freq_hz => REF_CLK_FREQ_HZ, -- natural := 200000000
    freq_meas_type  => FREQ_MEAS_TYPE)  -- freq_meas_type_t := FREQ_IN_HZ)
  port map(
    ref_clk       => pll_ref_clk,          -- in    std_logic
    smp_clk       => pll_ref_clk,          -- in    std_logic
    read_clk      => ocp_clk,              -- in    std_logic
    read_freq     => pll_ref_clk_freq,     -- out   std_logic_vector(31 downto 0)
    read_update   => pll_ref_clk_update,   -- out   std_logic
    read_running  => pll_ref_clk_running); -- out   std_logic)

  -- Measure pll_reg_clk frequency
  -- Use pll_ref_clk as reference clock frequency
  -- Use ocp_clk as read clock frequency
  pll_reg_clk_clock_speed_i : clock_speed
  generic map(
    ref_clk_freq_hz => REF_CLK_FREQ_HZ, -- natural := 200000000
    freq_meas_type  => FREQ_MEAS_TYPE)  -- freq_meas_type_t := FREQ_IN_HZ)
  port map(
    ref_clk       => pll_ref_clk,          -- in    std_logic
    smp_clk       => pll_reg_clk,          -- in    std_logic
    read_clk      => ocp_clk,              -- in    std_logic
    read_freq     => pll_reg_clk_freq,     -- out   std_logic_vector(31 downto 0)
    read_update   => pll_reg_clk_update,   -- out   std_logic
    read_running  => pll_reg_clk_running); -- out   std_logic)

  -- Measure pll_mem_clk frequency
  -- Use pll_ref_clk as reference clock frequency
  -- Use ocp_clk as read clock frequency
  pll_mem_clk_clock_speed_i : clock_speed
  generic map(
    ref_clk_freq_hz => REF_CLK_FREQ_HZ, -- natural := 200000000
    freq_meas_type  => FREQ_MEAS_TYPE)  -- freq_meas_type_t := FREQ_IN_HZ)
  port map(
    ref_clk       => pll_ref_clk,          -- in    std_logic
    smp_clk       => pll_mem_clk,          -- in    std_logic
    read_clk      => ocp_clk,              -- in    std_logic
    read_freq     => pll_mem_clk_freq,     -- out   std_logic_vector(31 downto 0)
    read_update   => pll_mem_clk_update,   -- out   std_logic
    read_running  => pll_mem_clk_running); -- out   std_logic)


  clk_vec_clock_speed_loop_g : for n in LCLK_NUM to MGT118_CLK1_NUM generate
    clk_vec_clock_speed_g : if ((BOARD_TYPE = ADM_XRC_6TL)and(CLK_EN_6TL(n) = '1'))or
                               ((BOARD_TYPE = ADM_XRC_6T1)and(CLK_EN_6T1(n) = '1'))generate
      -- Measure clk_vec frequencies
      -- Use pll_ref_clk as reference clock frequency
      -- Use ocp_clk as read clock frequency
      clk_vec_clock_speed_i : clock_speed
      generic map(
        ref_clk_freq_hz => REF_CLK_FREQ_HZ,  -- natural := 200000000
        freq_meas_type  => FREQ_MEAS_TYPE)   -- freq_meas_type_t := FREQ_IN_HZ)
      port map(
        ref_clk       => pll_ref_clk,       -- in    std_logic
        smp_clk       => clk_vec(n),        -- in    std_logic
        read_clk      => ocp_clk,           -- in    std_logic
        read_freq     => clk_freq(n),       -- out   std_logic_vector(31 downto 0)
        read_update   => clk_update(n),     -- out   std_logic
        read_running  => clk_running(n));   -- out   std_logic)
    end generate;
  end generate;


  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width,   -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH)    -- natural range 8 to ADB3_OCP_DATA_WIDTH
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      ctrl_reg    <= (others => '0');
      ctrl_reg_wr <= (others => '0');
      sel_reg     <= (others => '0');
    elsif rising_edge(ocp_clk) then
      ctrl_reg_wr <= (others => '0');
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          if    (ds_a = CLK_READ_CTRL_REG_ADDR(addr_width-1 downto 0)) then
            ctrl_reg(u downto l) <= ds_d(u downto l);
            ctrl_reg_wr(i) <= '1';
          elsif (ds_a = CLK_READ_SEL_REG_ADDR(addr_width-1 downto 0)) then
            sel_reg(u downto l) <= ds_d(u downto l);
          end if;
        end if;
      end loop;
    end if;
  end process;

  ctrl_clr_update <= ctrl_reg(31);          -- clear updates
  ctrl_wr         <= AND_reduce(ctrl_reg_wr);

  sel_clk         <= sel_reg(4 downto 0);   -- clock select bits


  -- Multiplex read registers
  rd_regs_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      ds_q <= (others => '0');
    elsif rising_edge(ocp_clk) then
      if    (ds_a = CLK_READ_SEL_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= sel_reg;
      elsif (ds_a = CLK_READ_STAT_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= stat_reg;
      elsif (ds_a = CLK_READ_FREQ_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= freq_reg;
      else
        ds_q <= (others => '-');
      end if;
    end if;
  end process;


  -- Generate rdata_reg by muxing clock freq counts
  clk_freq_p : process(ocp_rst,ocp_clk)
    variable pll_reg_clk_stat : std_logic_vector(2 downto 0);
    variable pll_ref_clk_stat : std_logic_vector(2 downto 0);
    variable pll_mem_clk_stat : std_logic_vector(2 downto 0);
    variable clk_freq_stat    : clk_vec_stat_t := (others => "000");
  begin
    if (ocp_rst = '1') then
      pll_reg_clk_update_sr <= '0';
      pll_ref_clk_update_sr <= '0';
      pll_mem_clk_update_sr <= '0';
      clk_update_sr         <= (others => '0');
    elsif rising_edge(ocp_clk) then
      -- Fix clock valid for board to '1'
      pll_reg_clk_stat := pll_reg_clk_update_sr & '1' & pll_reg_clk_running;
      pll_ref_clk_stat := pll_ref_clk_update_sr & '1' & pll_ref_clk_running;
      pll_mem_clk_stat := pll_mem_clk_update_sr & '1' & pll_mem_clk_running;

      -- Generate clock valid for board from CLK_EN_###(n)
      stat_comb_l : for n in LCLK_NUM to MGT118_CLK1_NUM loop
        if (BOARD_TYPE = ADM_XRC_6TL) then
          clk_freq_stat(n) := clk_update_sr(n) & CLK_EN_6TL(n) & clk_running(n);
        end if;
        if (BOARD_TYPE = ADM_XRC_6T1) then
          clk_freq_stat(n) := clk_update_sr(n) & CLK_EN_6T1(n) & clk_running(n);
        end if;
      end loop;

      case sel_clk is
        -- Internal clocks
        when PLL_REG_CLK_SEL => stat_reg(31 downto 29) <= pll_reg_clk_stat;
        when PLL_REF_CLK_SEL => stat_reg(31 downto 29) <= pll_ref_clk_stat;
        when PLL_MEM_CLK_SEL => stat_reg(31 downto 29) <= pll_mem_clk_stat;
        -- ...
        -- Non MGT external cloc
        when LCLK_SEL        => stat_reg(31 downto 29) <= clk_freq_stat(LCLK_NUM);
        when XRM_CLKIN_SEL   => stat_reg(31 downto 29) <= clk_freq_stat(XRM_CLKIN_NUM);

        -- MGT external clocks
        when MGT112_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT112_CLK0_NUM);
        when MGT113_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT113_CLK0_NUM);
        when MGT113_CLK1_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT113_CLK1_NUM);
        when MGT114_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT114_CLK0_NUM);
        when MGT115_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT115_CLK0_NUM);
        when MGT116_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT116_CLK0_NUM);
        when MGT117_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT117_CLK0_NUM);
        when others          => stat_reg <= (others => '0');
      end case;

      case sel_clk is
        -- Internal clocks
        when PLL_REG_CLK_SEL => freq_reg <= pll_reg_clk_freq;
        when PLL_REF_CLK_SEL => freq_reg <= pll_ref_clk_freq;
        when PLL_MEM_CLK_SEL => freq_reg <= pll_mem_clk_freq;
        -- ...
        -- Non MGT external cloc
        when LCLK_SEL        => freq_reg <= clk_freq(LCLK_NUM);
        when XRM_CLKIN_SEL   => freq_reg <= clk_freq(XRM_CLKIN_NUM);

        -- MGT external clocks
        when MGT112_CLK0_SEL => freq_reg <= clk_freq(MGT112_CLK0_NUM);
        when MGT113_CLK0_SEL => freq_reg <= clk_freq(MGT113_CLK0_NUM);
        when MGT113_CLK1_SEL => freq_reg <= clk_freq(MGT113_CLK1_NUM);
        when MGT114_CLK0_SEL => freq_reg <= clk_freq(MGT114_CLK0_NUM);
        when MGT115_CLK0_SEL => freq_reg <= clk_freq(MGT115_CLK0_NUM);
        when MGT116_CLK0_SEL => freq_reg <= clk_freq(MGT116_CLK0_NUM);
        when MGT117_CLK0_SEL => freq_reg <= clk_freq(MGT117_CLK0_NUM);
        when others          => freq_reg <= (others => '0');
      end case;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_ref_clk_update_sr <= '0';
      elsif (pll_ref_clk_update = '1') then
        pll_ref_clk_update_sr <= '1';
      end if;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_reg_clk_update_sr <= '0';
      elsif (pll_reg_clk_update = '1') then
        pll_reg_clk_update_sr <= '1';
      end if;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_mem_clk_update_sr <= '0';
      elsif (pll_mem_clk_update = '1') then
        pll_mem_clk_update_sr <= '1';
      end if;

      update_sr_l : for n in LCLK_NUM to MGT118_CLK1_NUM loop
        if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
          clk_update_sr(n) <= '0';
        elsif (clk_update(n) = '1') then
          clk_update_sr(n) <= '1';
        end if;
      end loop;
    end if;
  end process;

end;