--==============================================================================
--
--     File Name: blk_ds_io_test.vhd
--       Version: 0.1
--          Date: 20 Apr 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Front/Rear IO test
--
--   Description: Front (XRM 146)/Rear (PN4/PN6) IO test
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 20 Apr 2010: Initial design
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.uber_pkg.all;

entity blk_ds_io_test is
generic (
  addr_width : natural range 1 to DS_ADDR_WIDTH);
port(
  -- OCP Interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- IO Interface
  xrm_gpio         : inout xrm_gpio_t;
  pn4_gpio         : inout pn4_gpio_t;
  pn6_gpio         : inout pn6_gpio_t);
end;


architecture synth of blk_ds_io_test is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- XRM GPIO signals
  -- Initialise tri_reg so that output enables are off
  signal xrm_gpio_da_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) ;
  signal xrm_gpio_db_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) ;
  signal xrm_gpio_dc_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) ;
  signal xrm_gpio_dd_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) ;
  signal xrm_gpio_cs_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) ;

  signal xrm_gpio_da_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_db_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dc_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dd_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_cs_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);

  signal xrm_gpio_tri : xrm_gpio_t;
  signal xrm_gpio_in  : xrm_gpio_t;
  signal xrm_gpio_out : xrm_gpio_t;

  -- PN4 GPIO signals
  -- Initialise tri_reg so that output enables are off
  signal pn4_gpio_p_tri_reg    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_n_tri_reg    : std_logic_vector(REG_WIDTH-1 downto 0);

  signal pn4_gpio_p_wdata_reg  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_n_wdata_reg  : std_logic_vector(REG_WIDTH-1 downto 0);

  signal pn4_gpio_tri : pn4_gpio_t;
  signal pn4_gpio_in  : pn4_gpio_t;
  signal pn4_gpio_out : pn4_gpio_t;

  -- PN6 GPIO signals
  -- Initialise tri_reg so that output enables are off
  signal pn6_gpio_ms_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn6_gpio_ls_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0);

  signal pn6_gpio_ms_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn6_gpio_ls_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);

  signal pn6_gpio_tri_reg      : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_wdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_rdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);

  signal pn6_gpio_tri : pn6_gpio_t;
  signal pn6_gpio_in  : pn6_gpio_t;
  signal pn6_gpio_out : pn6_gpio_t;

begin

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width,   -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH)    -- natural range 8 to ADB3_OCP_DATA_WIDTH
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      xrm_gpio_da_tri_reg <= (others => '1');
      xrm_gpio_db_tri_reg <= (others => '1');
      xrm_gpio_dc_tri_reg <= (others => '1');
      xrm_gpio_dd_tri_reg <= (others => '1');
      xrm_gpio_cs_tri_reg <= (others => '1');
      pn4_gpio_p_tri_reg  <= (others => '1');
      pn4_gpio_n_tri_reg  <= (others => '1');
      pn6_gpio_ms_tri_reg <= (others => '1');
      pn6_gpio_ls_tri_reg <= (others => '1');
    elsif rising_edge(ocp_clk) then
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          if    (ds_a = XRM_GPIO_DA_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_da_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DA_DATA_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_da_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DB_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_db_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DB_DATA_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_db_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DC_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dc_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DC_DATA_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dc_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DD_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dd_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_DD_DATA_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dd_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_CS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_cs_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = XRM_GPIO_CS_DATA_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_cs_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN4_GPIO_P_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_p_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN4_GPIO_N_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_n_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN4_GPIO_P_DATA_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_p_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN4_GPIO_N_DATA_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_n_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_ms_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_ls_tri_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN6_GPIO_MS_DATA_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_ms_wdata_reg(u downto l) <= ds_d(u downto l);
          elsif (ds_a = PN6_GPIO_LS_DATA_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_ls_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
        end if;
      end loop;
    end if;
  end process;

  -- Connect XRM GPIO data tristates
  xrm_gpio_tri.xrm_da_p <= xrm_gpio_da_tri_reg(31 downto 16);
  xrm_gpio_tri.xrm_da_n <= xrm_gpio_da_tri_reg(15 downto 0);
  xrm_gpio_tri.xrm_db_p <= xrm_gpio_db_tri_reg(31 downto 16);
  xrm_gpio_tri.xrm_db_n <= xrm_gpio_db_tri_reg(15 downto 0);
  xrm_gpio_tri.xrm_dc_p <= xrm_gpio_dc_tri_reg(31 downto 16);
  xrm_gpio_tri.xrm_dc_n <= xrm_gpio_dc_tri_reg(15 downto 0);
  xrm_gpio_tri.xrm_dd_p <= xrm_gpio_dd_tri_reg(31 downto 16);
  xrm_gpio_tri.xrm_dd_n <= xrm_gpio_dd_tri_reg(15 downto 0);

  -- Connect XRM GPIO data outputs
  xrm_gpio_out.xrm_da_p <= xrm_gpio_da_wdata_reg(31 downto 16);
  xrm_gpio_out.xrm_da_n <= xrm_gpio_da_wdata_reg(15 downto 0);
  xrm_gpio_out.xrm_db_p <= xrm_gpio_db_wdata_reg(31 downto 16);
  xrm_gpio_out.xrm_db_n <= xrm_gpio_db_wdata_reg(15 downto 0);
  xrm_gpio_out.xrm_dc_p <= xrm_gpio_dc_wdata_reg(31 downto 16);
  xrm_gpio_out.xrm_dc_n <= xrm_gpio_dc_wdata_reg(15 downto 0);
  xrm_gpio_out.xrm_dd_p <= xrm_gpio_dd_wdata_reg(31 downto 16);
  xrm_gpio_out.xrm_dd_n <= xrm_gpio_dd_wdata_reg(15 downto 0);

  -- Connect XRM GPIO data tristates
  xrm_gpio_tri.xrm_sa <= xrm_gpio_cs_tri_reg(1 downto 0);
  xrm_gpio_tri.xrm_sb <= xrm_gpio_cs_tri_reg(3 downto 2);
  xrm_gpio_tri.xrm_sc <= xrm_gpio_cs_tri_reg(5 downto 4);
  xrm_gpio_tri.xrm_sd <= xrm_gpio_cs_tri_reg(9 downto 6);

  -- Connect XRM GPIO data outputs
  xrm_gpio_out.xrm_sa <= xrm_gpio_cs_wdata_reg(1 downto 0);
  xrm_gpio_out.xrm_sb <= xrm_gpio_cs_wdata_reg(3 downto 2);
  xrm_gpio_out.xrm_sc <= xrm_gpio_cs_wdata_reg(5 downto 4);
  xrm_gpio_out.xrm_sd <= xrm_gpio_cs_wdata_reg(9 downto 6);

  -- Connect XRM GPIO data tristates
  xrm_gpio_tri.xrm_da_cc_n <= xrm_gpio_cs_tri_reg(10);
  xrm_gpio_tri.xrm_da_cc_p <= xrm_gpio_cs_tri_reg(11);
  xrm_gpio_tri.xrm_db_cc_n <= xrm_gpio_cs_tri_reg(12);
  xrm_gpio_tri.xrm_db_cc_p <= xrm_gpio_cs_tri_reg(13);
  xrm_gpio_tri.xrm_dc_cc_n <= xrm_gpio_cs_tri_reg(14);
  xrm_gpio_tri.xrm_dc_cc_p <= xrm_gpio_cs_tri_reg(15);
  xrm_gpio_tri.xrm_dd_cc_n <= xrm_gpio_cs_tri_reg(16);
  xrm_gpio_tri.xrm_dd_cc_p <= xrm_gpio_cs_tri_reg(17);

  -- Connect XRM GPIO data outputs
  xrm_gpio_out.xrm_da_cc_n <= xrm_gpio_cs_wdata_reg(10);
  xrm_gpio_out.xrm_da_cc_p <= xrm_gpio_cs_wdata_reg(11);
  xrm_gpio_out.xrm_db_cc_n <= xrm_gpio_cs_wdata_reg(12);
  xrm_gpio_out.xrm_db_cc_p <= xrm_gpio_cs_wdata_reg(13);
  xrm_gpio_out.xrm_dc_cc_n <= xrm_gpio_cs_wdata_reg(14);
  xrm_gpio_out.xrm_dc_cc_p <= xrm_gpio_cs_wdata_reg(15);
  xrm_gpio_out.xrm_dd_cc_n <= xrm_gpio_cs_wdata_reg(16);
  xrm_gpio_out.xrm_dd_cc_p <= xrm_gpio_cs_wdata_reg(17);

  -- Connect PN4 GPIO data tristates
  pn4_gpio_tri.gpio_p <= pn4_gpio_p_tri_reg(PN4_GPIO_WIDTH-1 downto 0);
  pn4_gpio_tri.gpio_n <= pn4_gpio_n_tri_reg(PN4_GPIO_WIDTH-1 downto 0);

  -- Connect PN4 GPIO data outputs
  pn4_gpio_out.gpio_p <= pn4_gpio_p_wdata_reg(PN4_GPIO_WIDTH-1 downto 0);
  pn4_gpio_out.gpio_n <= pn4_gpio_n_wdata_reg(PN4_GPIO_WIDTH-1 downto 0);

  -- Connect PN6 GPIO data tristates
  pn6_gpio_tri_reg    <= pn6_gpio_ms_tri_reg & pn6_gpio_ls_tri_reg;
  pn6_gpio_tri.gpio   <= pn6_gpio_tri_reg(PN6_GPIO_WIDTH-1 downto 0);

  -- Connect PN6 GPIO data outputs
  pn6_gpio_wdata_reg  <= pn6_gpio_ms_wdata_reg & pn6_gpio_ls_wdata_reg;
  pn6_gpio_out.gpio   <= pn6_gpio_wdata_reg(PN6_GPIO_WIDTH-1 downto 0);

  -- Connect PN6 GPIO data inputs
  pn6_gpio_rdata_reg(PN6_GPIO_WIDTH-1 downto 0) <= pn6_gpio_in.gpio;


  -- Multiplex read registers
  rd_regs_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      ds_q <= (others => '0');
    elsif rising_edge(ocp_clk) then
      -- XRM read registers
      if    (ds_a = XRM_GPIO_DA_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_da_tri_reg;
      elsif (ds_a = XRM_GPIO_DA_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_in.xrm_da_p & xrm_gpio_in.xrm_da_n;
      elsif (ds_a = XRM_GPIO_DB_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_db_tri_reg;
      elsif (ds_a = XRM_GPIO_DB_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_in.xrm_db_p & xrm_gpio_in.xrm_db_n;
      elsif (ds_a = XRM_GPIO_DC_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_dc_tri_reg;
      elsif (ds_a = XRM_GPIO_DC_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_in.xrm_dc_p & xrm_gpio_in.xrm_dc_n;
      elsif (ds_a = XRM_GPIO_DD_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_dd_tri_reg;
      elsif (ds_a = XRM_GPIO_DD_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_in.xrm_dd_p & xrm_gpio_in.xrm_dd_n;
      elsif (ds_a = XRM_GPIO_CS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= xrm_gpio_cs_tri_reg;
      elsif (ds_a = XRM_GPIO_CS_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q(17 downto 0) <= xrm_gpio_in.xrm_dd_cc_p & xrm_gpio_in.xrm_dd_cc_n &
                             xrm_gpio_in.xrm_dc_cc_p & xrm_gpio_in.xrm_dc_cc_n &
                             xrm_gpio_in.xrm_db_cc_p & xrm_gpio_in.xrm_db_cc_n &
                             xrm_gpio_in.xrm_da_cc_p & xrm_gpio_in.xrm_da_cc_n &
                             xrm_gpio_in.xrm_sd & xrm_gpio_in.xrm_sc & xrm_gpio_in.xrm_sb & xrm_gpio_in.xrm_sa;
      -- PN4 read registers
      elsif (ds_a = PN4_GPIO_P_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q(PN4_GPIO_WIDTH-1 downto 0) <= pn4_gpio_p_tri_reg(PN4_GPIO_WIDTH-1 downto 0);
      elsif (ds_a = PN4_GPIO_P_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q(PN4_GPIO_WIDTH-1 downto 0) <= pn4_gpio_in.gpio_p;
      elsif (ds_a = PN4_GPIO_N_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q(PN4_GPIO_WIDTH-1 downto 0) <= pn4_gpio_n_tri_reg(PN4_GPIO_WIDTH-1 downto 0);
      elsif (ds_a = PN4_GPIO_N_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q(PN4_GPIO_WIDTH-1 downto 0) <= pn4_gpio_in.gpio_n;
      -- PN6 read registers
      elsif (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= pn6_gpio_ms_tri_reg;
      elsif (ds_a = PN6_GPIO_MS_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= pn6_gpio_rdata_reg(63 downto 32);
      elsif (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= pn6_gpio_ls_tri_reg;
      elsif (ds_a = PN6_GPIO_LS_DATA_REG_ADDR(addr_width-1 downto 0)) then
        ds_q <= pn6_gpio_rdata_reg(31 downto 0);
      else
        ds_q <= (others => '-');
      end if;
    end if;
  end process;


  -- XRM Bidirectional GPIO
  gen_xrm_d_g : for n in 0 to 15 generate
    xrm_gpio.xrm_da_p(n) <= xrm_gpio_out.xrm_da_p(n) when (xrm_gpio_tri.xrm_da_p(n) = '0') else 'Z';
    xrm_gpio.xrm_da_n(n) <= xrm_gpio_out.xrm_da_n(n) when (xrm_gpio_tri.xrm_da_n(n) = '0') else 'Z';
    xrm_gpio.xrm_db_p(n) <= xrm_gpio_out.xrm_db_p(n) when (xrm_gpio_tri.xrm_db_p(n) = '0') else 'Z';
    xrm_gpio.xrm_db_n(n) <= xrm_gpio_out.xrm_db_n(n) when (xrm_gpio_tri.xrm_db_n(n) = '0') else 'Z';
    xrm_gpio.xrm_dc_p(n) <= xrm_gpio_out.xrm_dc_p(n) when (xrm_gpio_tri.xrm_dc_p(n) = '0') else 'Z';
    xrm_gpio.xrm_dc_n(n) <= xrm_gpio_out.xrm_dc_n(n) when (xrm_gpio_tri.xrm_dc_n(n) = '0') else 'Z';
    xrm_gpio.xrm_dd_p(n) <= xrm_gpio_out.xrm_dd_p(n) when (xrm_gpio_tri.xrm_dd_p(n) = '0') else 'Z';
    xrm_gpio.xrm_dd_n(n) <= xrm_gpio_out.xrm_dd_n(n) when (xrm_gpio_tri.xrm_dd_n(n) = '0') else 'Z';
    xrm_gpio_in.xrm_da_p(n) <= xrm_gpio.xrm_da_p(n);
    xrm_gpio_in.xrm_da_n(n) <= xrm_gpio.xrm_da_n(n);
    xrm_gpio_in.xrm_db_p(n) <= xrm_gpio.xrm_db_p(n);
    xrm_gpio_in.xrm_db_n(n) <= xrm_gpio.xrm_db_n(n);
    xrm_gpio_in.xrm_dc_p(n) <= xrm_gpio.xrm_dc_p(n);
    xrm_gpio_in.xrm_dc_n(n) <= xrm_gpio.xrm_dc_n(n);
    xrm_gpio_in.xrm_dd_p(n) <= xrm_gpio.xrm_dd_p(n);
    xrm_gpio_in.xrm_dd_n(n) <= xrm_gpio.xrm_dd_n(n);
  end generate;

  gen_xrm_s_g : for n in 0 to 1 generate
    xrm_gpio.xrm_sa(n) <= xrm_gpio_out.xrm_sa(n) when (xrm_gpio_tri.xrm_sa(n) = '0') else 'Z';
    xrm_gpio.xrm_sb(n) <= xrm_gpio_out.xrm_sb(n) when (xrm_gpio_tri.xrm_sb(n) = '0') else 'Z';
    xrm_gpio.xrm_sc(n) <= xrm_gpio_out.xrm_sc(n) when (xrm_gpio_tri.xrm_sc(n) = '0') else 'Z';
    xrm_gpio_in.xrm_sa(n) <= xrm_gpio.xrm_sa(n);
    xrm_gpio_in.xrm_sb(n) <= xrm_gpio.xrm_sb(n);
    xrm_gpio_in.xrm_sc(n) <= xrm_gpio.xrm_sc(n);
  end generate;

  gen_xrm_sd_g : for n in 0 to 3 generate
    xrm_gpio.xrm_sd(n) <= xrm_gpio_out.xrm_sd(n) when (xrm_gpio_tri.xrm_sd(n) = '0') else 'Z';
    xrm_gpio_in.xrm_sd(n) <= xrm_gpio.xrm_sd(n);
  end generate;

  xrm_gpio.xrm_da_cc_p <= xrm_gpio_out.xrm_da_cc_p when (xrm_gpio_tri.xrm_da_cc_p = '0') else 'Z';
  xrm_gpio.xrm_da_cc_n <= xrm_gpio_out.xrm_da_cc_n when (xrm_gpio_tri.xrm_da_cc_n = '0') else 'Z';
  xrm_gpio.xrm_db_cc_p <= xrm_gpio_out.xrm_db_cc_p when (xrm_gpio_tri.xrm_db_cc_p = '0') else 'Z';
  xrm_gpio.xrm_db_cc_n <= xrm_gpio_out.xrm_db_cc_n when (xrm_gpio_tri.xrm_db_cc_n = '0') else 'Z';
  xrm_gpio.xrm_dc_cc_p <= xrm_gpio_out.xrm_dc_cc_p when (xrm_gpio_tri.xrm_dc_cc_p = '0') else 'Z';
  xrm_gpio.xrm_dc_cc_n <= xrm_gpio_out.xrm_dc_cc_n when (xrm_gpio_tri.xrm_dc_cc_n = '0') else 'Z';
  xrm_gpio.xrm_dd_cc_p <= xrm_gpio_out.xrm_dd_cc_p when (xrm_gpio_tri.xrm_dd_cc_p = '0') else 'Z';
  xrm_gpio.xrm_dd_cc_n <= xrm_gpio_out.xrm_dd_cc_n when (xrm_gpio_tri.xrm_dd_cc_n = '0') else 'Z';
  xrm_gpio_in.xrm_da_cc_p <= xrm_gpio.xrm_da_cc_p;
  xrm_gpio_in.xrm_da_cc_n <= xrm_gpio.xrm_da_cc_n;
  xrm_gpio_in.xrm_db_cc_p <= xrm_gpio.xrm_db_cc_p;
  xrm_gpio_in.xrm_db_cc_n <= xrm_gpio.xrm_db_cc_n;
  xrm_gpio_in.xrm_dc_cc_p <= xrm_gpio.xrm_dc_cc_p;
  xrm_gpio_in.xrm_dc_cc_n <= xrm_gpio.xrm_dc_cc_n;
  xrm_gpio_in.xrm_dd_cc_p <= xrm_gpio.xrm_dd_cc_p;
  xrm_gpio_in.xrm_dd_cc_n <= xrm_gpio.xrm_dd_cc_n;

  -- PN4 Bidirectional GPIO
  gen_pn4_g : for n in 1 to PN4_GPIO_WIDTH generate
    pn4_gpio.gpio_p(n)    <= pn4_gpio_out.gpio_p(n) when (pn4_gpio_tri.gpio_p(n) = '0') else 'Z';
    pn4_gpio.gpio_n(n)    <= pn4_gpio_out.gpio_n(n) when (pn4_gpio_tri.gpio_n(n) = '0') else 'Z';
    pn4_gpio_in.gpio_p(n) <= pn4_gpio.gpio_p(n);
    pn4_gpio_in.gpio_n(n) <= pn4_gpio.gpio_n(n);
  end generate;

  -- PN6 Bidirectional GPIO
  gen_pn6_g : for n in 1 to PN6_GPIO_WIDTH generate
    pn6_gpio.gpio(n)      <= pn6_gpio_out.gpio(n) when (pn6_gpio_tri.gpio(n) = '0')  else 'Z';
    pn6_gpio_in.gpio(n)   <= pn6_gpio.gpio(n);
  end generate;

end;