--==============================================================================
--
--     File Name: uber_tb_pkg.vhd
--       Version: 0.1
--          Date: 25 Mar 2010
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package uber_tb_pkg
--
--   Description: Testbench package for uber target FPGA
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Mar 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_pkg.all;
use work.uber_pkg.all;

package uber_tb_pkg is

  -- Fixed Clock periods
  constant CLK_250_PERIOD : time :=  4.00 ns;    -- 250 MHz
  constant CLK_200_PERIOD : time :=  5.00 ns;    -- 200 MHz
  constant CLK_166_PERIOD : time :=  6.00 ns;    -- 166 MHz
  constant CLK_125_PERIOD : time :=  8.00 ns;    -- 125 MHz
  constant CLK_100_PERIOD : time := 10.00 ns;    -- 125 MHz
  constant CLK_80_PERIOD  : time := 12.50 ns;    --  80 MHz

  -- Clock frequencies in MHz (truncated division)
  constant CLK_250_FREQ   : natural := 1 us / CLK_250_PERIOD;
  constant CLK_200_FREQ   : natural := 1 us / CLK_200_PERIOD;
  constant CLK_166_FREQ   : natural := 1 us / CLK_166_PERIOD;
  constant CLK_125_FREQ   : natural := 1 us / CLK_125_PERIOD;
  constant CLK_100_FREQ   : natural := 1 us / CLK_100_PERIOD;
  constant CLK_80_FREQ    : natural := 1 us / CLK_80_PERIOD;

  -- Define DMA OCP channel function
  constant DMA_WRITE_CHANNEL : natural range 0 to MAX_DMA_CHANNELS-1 := 0;
  constant DMA_READ_CHANNEL  : natural range 0 to MAX_DMA_CHANNELS-1 := 1;
  constant DMA_DATA_WIDTH    : natural range 8 to ADB3_OCP_DATA_WIDTH := 128;

  -- Constants for dma channel 0/1 test
  constant DMA_SIZE : natural := 4032;

  -- Select Direct Master inclusion
  type master_use_t is (MASTER_ON,MASTER_OFF);

  constant MASTER_USE : master_use_t := MASTER_OFF;
  constant DM_CHANNEL : natural range 0 to MAX_DM_CHANNELS-1 := 0;

  -- Constants for testbench direct master register interface to simulate host memory
  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant HOST_MASTER_REG_NUM  : natural := 5;
  constant HOST_MASTER_REG_OFF  : natural := HOST_MASTER_REG_NUM * (DM_DATA_WIDTH/8);
  constant HOST_MASTER_REG_ADDR : dm_addr_s := std_logic_vector(TO_UNSIGNED((HOST_MASTER_REG_OFF),DM_ADDR_WIDTH));

  -- Constants for IO
  constant XRM_GPIO_EN : std_logic_vector(3 downto 0) := "0001";
  constant PN4_GPIO_EN : std_logic_vector(3 downto 0) := "0010";
  constant PN6_GPIO_EN : std_logic_vector(3 downto 0) := "0100";
  constant IO_DISABLE  : std_logic_vector(3 downto 0) := "0000";

  -- Constants for interrupt
  constant MASK_EN_ALL : std_logic_vector(31 downto 0) := X"00000000";

  -- Constant to control OCP transaction probe file logging
  constant PROBE_LOG : boolean := false;

  -- Used in test_uber and test_uber_ds to indicate falling edge on linti_n
  shared variable linti_n_edge : std_logic := '0';

  -- Test status
  type top_comp_t is record
    mptl_gpio_complete : boolean;
    dma_abort_complete : boolean;
  end record;

  type ds_comp_t is record
    simple_complete    : boolean;
    master_complete    : boolean;
    clock_complete     : boolean;
    frontio_complete   : boolean;
    reario_complete    : boolean;
    int_complete       : boolean;
    info_complete      : boolean;
    bram_complete      : boolean;
  end record;

  type dma_comp_t is record
    dma_write_complete : boolean;
    dma_read_complete  : boolean;
  end record;

  type top_pass_t is record
    mptl_gpio_passed   : boolean;
    dma_abort_passed   : boolean;
  end record;

  type ds_pass_t is record
    simple_passed      : boolean;
    master_passed      : boolean;
    clock_passed       : boolean;
    frontio_passed     : boolean;
    reario_passed      : boolean;
    int_passed         : boolean;
    info_passed        : boolean;
    bram_passed        : boolean;
  end record;

  type dma_pass_t is record
    dma_write_passed   : boolean;
    dma_read_passed    : boolean;
  end record;

end;


package body uber_tb_pkg is

end;